# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
from __future__ import annotations

import numpy as np
from typing import Union, List, Set, Tuple, Dict
from PySide6 import QtGui, QtCore, QtWidgets
from matplotlib import pyplot as plt
from matplotlib.colors import LinearSegmentedColormap

from VeraGrid.Gui.associations_model import AssociationsModel
from VeraGrid.Gui.table_view_header_wrap import HeaderViewWithWordWrap
from VeraGridEngine.Compilers.circuit_to_data import compile_numerical_circuit_at
from VeraGrid.Gui.Main.SubClasses.Model.compiled_arrays_model import CompiledArraysModule
from VeraGridEngine.Topology.VoltageLevels import vl_creation_common_functions as substation_wizards
import VeraGridEngine.basic_structures as bs
import VeraGridEngine.Devices as dev
import VeraGrid.Gui.gui_functions as gf
from VeraGrid.Gui.object_model import ObjectsModel
from VeraGrid.Gui.object_proxy_model import ObjectModelFilterProxy
from VeraGrid.Gui.profiles_model import ProfilesModel
from VeraGridEngine.enumerations import DeviceType, TimeSeriesSearchPoint
from VeraGridEngine.Devices.types import ALL_DEV_TYPES
from VeraGridEngine.Topology.detect_substations import detect_substations, detect_facilities
from VeraGrid.Gui.Analysis.object_plot_analysis import object_histogram_analysis
from VeraGrid.Gui.messages import yes_no_question, warning_msg, info_msg
from VeraGrid.Gui.Main.SubClasses.Model.diagrams import DiagramsMain
from VeraGrid.Gui.Diagrams.Editors.line_editor import LineEditor
from VeraGrid.Gui.TowerBuilder.LineBuilderDialogue import TowerBuilderGUI
from VeraGrid.Gui.RmsModelEditor.rms_model_editor_engine import RmsModelEditorGUI
from VeraGrid.Gui.rms_events_editor_dialog import RmsEventEditor
from VeraGrid.Gui.SystemScaler.system_scaler import SystemScaler
from VeraGrid.Gui.Diagrams.MapWidget.grid_map_widget import GridMapWidget, generate_map_diagram
from VeraGrid.Gui.Diagrams.SchematicWidget.schematic_widget import SchematicWidget, make_diagram_from_buses
from VeraGrid.Gui.GridReduce.grid_reduce import GridReduceDialogue
from VeraGrid.Gui.SubstationDesigner.substation_designer import SubstationDesigner
from VeraGrid.Gui.general_dialogues import LogsDialogue, CustomQuestionDialogue, CheckListDialogue
from VeraGrid.Gui.Diagrams.Editors.transformer_editor import TransformerEditor
from VeraGrid.Gui.Diagrams.Editors.transformer3w_editor import Transformer3WEditor
from VeraGrid.Gui.Diagrams.Editors.controllable_shunt_editor import ControllableShuntEditor


class DataBaseTableMain(DiagramsMain):
    """
    Diagrams Main
    """

    def __init__(self, parent=None):
        """

        @param parent:
        """

        # create main window
        DiagramsMain.__init__(self, parent)

        # list of all the objects of the selected type under the Objects tab
        self.type_objects_list = list()

        # setup the objects tree
        self.setup_objects_tree()

        # setup the tree for compiled arrays
        self.setup_compiled_arrays_tree()

        (self.ts_search_points_dict,
         ts_search_points_mdl) = gf.enums_to_model(
            [TimeSeriesSearchPoint.HighestLoad,
             TimeSeriesSearchPoint.LowestLoad]
        )
        self.ui.goToTsPointComboBox.setModel(ts_search_points_mdl)

        self.ui.smart_search_lineEdit.setPlaceholderText("Type the object name or a smart filter expression ...")

        # Buttons
        self.ui.filter_pushButton.clicked.connect(self.objects_smart_search)
        self.ui.delete_selected_objects_pushButton.clicked.connect(self.delete_selected_db_table_objects)
        self.ui.add_object_pushButton.clicked.connect(self.add_objects)
        self.ui.structure_analysis_pushButton.clicked.connect(self.objects_histogram_analysis_plot)
        self.ui.goToTsPointButton.clicked.connect(self.search_ts_point)

        # menu trigger
        self.ui.actionDelete_inconsistencies.triggered.connect(self.delete_inconsistencies)
        self.ui.actionClean_database.triggered.connect(self.clean_database)
        self.ui.actionScale.triggered.connect(self.scale)
        self.ui.actionDetect_substations.triggered.connect(self.detect_substations)
        self.ui.actionDetect_facilities.triggered.connect(self.detect_facilities)
        self.ui.actionGrid_reduction.triggered.connect(self.grid_reduction_from_schematic_selection)
        self.ui.actionSubstation_wizard.triggered.connect(self.add_substation_with_wizard)

        # tree click
        self.ui.dataStructuresTreeView.clicked.connect(self.view_objects_data)

        # line edit enter
        self.ui.smart_search_lineEdit.returnPressed.connect(self.objects_smart_search)

        # context menu
        self.ui.dataStructureTableView.customContextMenuRequested.connect(self.show_objects_context_menu)

        # Set context menu policy to CustomContextMenu
        self.ui.dataStructureTableView.setContextMenuPolicy(QtGui.Qt.ContextMenuPolicy.CustomContextMenu)

        # wrap headers
        self.ui.dataStructureTableView.setHorizontalHeader(HeaderViewWithWordWrap(self.ui.dataStructureTableView))
        self.ui.profiles_tableView.setHorizontalHeader(HeaderViewWithWordWrap(self.ui.profiles_tableView))
        self.ui.associationsTableView.setHorizontalHeader(HeaderViewWithWordWrap(self.ui.associationsTableView))

        # combobox change
        self.ui.associationsComboBox.currentTextChanged.connect(self.on_associations_combo_box_change)

    def setup_objects_tree(self):
        """
        Setup the database left tree object
        """
        icons = {
            "Regions": ":/Icons/icons/map.png",
            DeviceType.CountryDevice.value: ":/Icons/icons/country.png",
            DeviceType.CommunityDevice.value: ":/Icons/icons/community.png",
            DeviceType.RegionDevice.value: ":/Icons/icons/region.png",
            DeviceType.MunicipalityDevice.value: ":/Icons/icons/municipality.png",
            DeviceType.AreaDevice.value: ":/Icons/icons/area.png",
            DeviceType.ZoneDevice.value: ":/Icons/icons/zone.png",

            "Substation": ":/Icons/icons/bus_icon.png",
            DeviceType.SubstationDevice.value: ":/Icons/icons/substation.png",
            DeviceType.VoltageLevelDevice.value: ":/Icons/icons/voltage_level.png",
            DeviceType.BusBarDevice.value: ":/Icons/icons/bus_bar_icon.png",
            DeviceType.BusDevice.value: ":/Icons/icons/bus_icon.png",
            DeviceType.SwitchDevice.value: ":/Icons/icons/switch.png",

            "Injections": ":/Icons/icons/add_load.png",
            DeviceType.GeneratorDevice.value: ":/Icons/icons/gen.png",
            DeviceType.BatteryDevice.value: ":/Icons/icons/batt.png",
            DeviceType.LoadDevice.value: ":/Icons/icons/load_dev.png",
            DeviceType.StaticGeneratorDevice.value: ":/Icons/icons/sta_gen.png",
            DeviceType.ExternalGridDevice.value: ":/Icons/icons/external_grid.png",
            DeviceType.ShuntDevice.value: ":/Icons/icons/shunt.png",
            DeviceType.ControllableShuntDevice.value: ":/Icons/icons/controllable_shunt.png",
            DeviceType.CurrentInjectionDevice.value: ":/Icons/icons/load_dev.png",

            "Branches": ":/Icons/icons/reactance.png",
            DeviceType.LineDevice.value: ":/Icons/icons/ac_line.png",
            DeviceType.DCLineDevice.value: ":/Icons/icons/dc.png",
            DeviceType.Transformer2WDevice.value: ":/Icons/icons/to_transformer.png",
            DeviceType.WindingDevice.value: ":/Icons/icons/winding.png",
            DeviceType.Transformer3WDevice.value: ":/Icons/icons/transformer3w.png",
            DeviceType.SeriesReactanceDevice.value: ":/Icons/icons/reactance.png",
            DeviceType.HVDCLineDevice.value: ":/Icons/icons/to_hvdc.png",
            DeviceType.VscDevice.value: ":/Icons/icons/vsc.png",
            DeviceType.UpfcDevice.value: ":/Icons/icons/upfc.png",

            "Fluid": ":/Icons/icons/dam_gray.png",
            DeviceType.FluidNodeDevice.value: ":/Icons/icons/fluid_node.png",
            DeviceType.FluidPathDevice.value: ":/Icons/icons/fluid_path.png",
            DeviceType.FluidTurbineDevice.value: ":/Icons/icons/fluid_turbine.png",
            DeviceType.FluidPumpDevice.value: ":/Icons/icons/fluid_pump.png",
            DeviceType.FluidP2XDevice.value: ":/Icons/icons/fluid_p2x.png",

            "Groups": ":/Icons/icons/groups.png",
            DeviceType.BranchGroupDevice.value: ":/Icons/icons/branch_group.png",
            DeviceType.ModellingAuthority.value: ":/Icons/icons/modelling_authority.png",
            DeviceType.FacilityDevice.value: ":/Icons/icons/powerplant.png",

            "Contingencies": ":/Icons/icons/contingency_group.png",
            DeviceType.ContingencyGroupDevice.value: ":/Icons/icons/contingency_group.png",
            DeviceType.ContingencyDevice.value: ":/Icons/icons/contingency.png",
            DeviceType.RemedialActionGroupDevice.value: ":/Icons/icons/remedial_action_group.png",
            DeviceType.RemedialActionDevice.value: ":/Icons/icons/remedial_action.png",
            DeviceType.ShortCircuitEvent.value: ":/Icons/icons/contingency.png",

            "Investments": ":/Icons/icons/investment_group.png",
            DeviceType.InvestmentsGroupDevice.value: ":/Icons/icons/investment_group.png",
            DeviceType.InvestmentDevice.value: ":/Icons/icons/investment_dev.png",

            "Dynamic": ":/Icons/icons/dyn_gray.png",
            DeviceType.RmsEventsGroupDevice.value: ":/Icons/icons/dyn_gray.png",
            DeviceType.RmsEventDevice.value: ":/Icons/icons/dyn_gray.png",

            "Associations": ":/Icons/icons/associations.png",
            DeviceType.Technology.value: ":/Icons/icons/technology.png",
            DeviceType.FuelDevice.value: ":/Icons/icons/fuel.png",
            DeviceType.EmissionGasDevice.value: ":/Icons/icons/emission.png",
            DeviceType.Owner.value: ":/Icons/icons/owner.png",

            "Measurements": ":/Icons/icons/measurement.png",
            DeviceType.PMeasurementDevice.value: ":/Icons/icons/measurement.png",
            DeviceType.QMeasurementDevice.value: ":/Icons/icons/measurement.png",
            DeviceType.PfMeasurementDevice.value: ":/Icons/icons/measurement.png",
            DeviceType.QfMeasurementDevice.value: ":/Icons/icons/measurement.png",
            DeviceType.PtMeasurementDevice.value: ":/Icons/icons/measurement.png",
            DeviceType.QtMeasurementDevice.value: ":/Icons/icons/measurement.png",
            DeviceType.PgMeasurementDevice.value: ":/Icons/icons/measurement.png",
            DeviceType.QgMeasurementDevice.value: ":/Icons/icons/measurement.png",
            DeviceType.IfMeasurementDevice.value: ":/Icons/icons/measurement.png",
            DeviceType.ItMeasurementDevice.value: ":/Icons/icons/measurement.png",
            DeviceType.VmMeasurementDevice.value: ":/Icons/icons/measurement.png",
            DeviceType.VaMeasurementDevice.value: ":/Icons/icons/measurement.png",

            "Catalogue": ":/Icons/icons/Catalogue.png",
            DeviceType.WireDevice.value: ":/Icons/icons/ac_line.png",
            DeviceType.OverheadLineTypeDevice.value: ":/Icons/icons/tower.png",
            DeviceType.UnderGroundLineDevice.value: ":/Icons/icons/ac_line.png",
            DeviceType.SequenceLineDevice.value: ":/Icons/icons/ac_line.png",
            DeviceType.TransformerTypeDevice.value: ":/Icons/icons/to_transformer.png",
            DeviceType.RmsModelTemplateDevice.value: ":/Icons/icons/dyn_gray.png",
            DeviceType.DynamicModelHostDevice.value: ":/Icons/icons/dyn_gray.png",

        }

        db_tree_model = gf.get_tree_model(d=self.circuit.get_template_objects_str_dict(),
                                          top='Objects',
                                          icons=icons)

        self.ui.dataStructuresTreeView.setModel(db_tree_model)
        self.ui.dataStructuresTreeView.setRootIsDecorated(True)
        self.expand_object_tree_nodes()

    def setup_compiled_arrays_tree(self):
        """

        :return:
        """
        mdl = gf.get_tree_model(d=CompiledArraysModule.available_structures,
                                top='Arrays')

        self.ui.simulationDataStructuresTreeView.setModel(mdl)

    def create_objects_model(self, elements, elm_type: DeviceType) -> ObjectsModel:
        """
        Generate the objects' table model
        :param elements: list of elements
        :param elm_type: name of DeviceType.BusDevice
        :return: QtCore.QAbstractTableModel
        """
        template_elm, dictionary_of_lists = self.circuit.get_dictionary_of_lists(elm_type=elm_type)

        mdl = ObjectsModel(objects=elements,
                           property_list=template_elm.property_list,
                           time_index=self.get_db_slider_index(),
                           parent=self.ui.dataStructureTableView,
                           editable=True,
                           dictionary_of_lists=dictionary_of_lists)

        return mdl

    def display_profiles(self, proxy_mdl: ObjectModelFilterProxy):
        """
        Display profile
        :param proxy_mdl: ObjectModelFilterProxy used for the object's table
        """
        if self.circuit.time_profile is not None:

            dev_type_text = self.get_db_object_selected_type()

            if dev_type_text is not None:

                magnitudes, mag_types = self.circuit.profile_magnitudes.get(dev_type_text, (list(), list()))

                if len(magnitudes) > 0:
                    # get the enumeration unique association with the device text
                    dev_type = self.circuit.device_type_name_dict[dev_type_text]

                    idx = self.ui.device_type_magnitude_comboBox.currentIndex()
                    magnitude = magnitudes[idx]
                    mtype = mag_types[idx]

                    mdl = ProfilesModel(time_array=self.circuit.get_time_array(),
                                        elements=proxy_mdl.objects,
                                        device_type=dev_type,
                                        magnitude=magnitude,
                                        data_format=mtype,
                                        parent=self.ui.profiles_tableView)
                    self.ui.profiles_tableView.setModel(mdl)
                else:
                    self.ui.profiles_tableView.setModel(None)
            else:
                self.ui.profiles_tableView.setModel(None)

    def display_associations(self, proxy_mdl: ObjectModelFilterProxy):
        """
        Display the association table
        :param proxy_mdl: ObjectModelFilterProxy used for the object's table
        :return:
        """
        dev_type_text = self.get_db_object_selected_type()
        association_property_name = self.ui.associationsComboBox.currentText()

        if dev_type_text is not None and proxy_mdl is not None and association_property_name != "":

            elements = proxy_mdl.objects

            if len(elements) > 0:

                gc_prop = elements[0].get_property_by_name(prop_name=association_property_name)
                associations: dev.Associations = elements[0].get_snapshot_value_by_name(name=association_property_name)
                associated_objects = self.circuit.get_elements_by_type(device_type=associations.device_type)
                self.ui.association_units_label.setText(gc_prop.units)

                if len(associated_objects) > 0:
                    mdl = AssociationsModel(objects=elements,
                                            associated_objects=associated_objects,
                                            gc_prop=gc_prop,
                                            table_view=self.ui.associationsTableView)

                    self.ui.associationsTableView.setModel(mdl)
                else:
                    self.ui.associationsTableView.setModel(None)
            else:
                self.ui.associationsTableView.setModel(None)
                self.ui.association_units_label.setText("")
        else:
            self.ui.associationsTableView.setModel(None)
            self.ui.association_units_label.setText("")

    def on_associations_combo_box_change(self):
        """
        Triggered on self.ui.associationsComboBox.currentTextChanged
        """
        self.display_associations(proxy_mdl=self.get_current_objects_model_view())

    def copy_objects_data(self):
        """
        Copy the current displayed objects table to the clipboard
        """
        mdl: ObjectModelFilterProxy | None = self.get_current_objects_model_view()
        if mdl is not None:
            mdl.copy_to_clipboard()
            self.show_info_toast('Copied!')
        else:
            warning_msg('There is no data displayed, please display one', 'Copy profile to clipboard')

    def get_db_object_selected_type(self) -> Union[None, str]:
        """
        Get the selected object type in the database tree view
        :return:
        """
        indices = self.ui.dataStructuresTreeView.selectedIndexes()

        if len(indices) > 0:
            return indices[0].data(role=QtCore.Qt.ItemDataRole.DisplayRole)
        else:
            return None

    def get_selected_objects_model(self) -> Tuple[ObjectsModel | None, List[ALL_DEV_TYPES] | None, str | None]:
        """
        Get the selected objects' model
        :return: ObjectsModel, list of objects, object type name
        """
        if len(self.ui.dataStructuresTreeView.selectedIndexes()) == 0:
            return None, None, None

        if self.ui.dataStructuresTreeView.selectedIndexes()[0].parent().row() > -1:
            # if the clicked element has a valid parent...

            elm_type = self.get_db_object_selected_type()

            if elm_type is not None:

                elements = self.circuit.get_elements_by_type(device_type=DeviceType(elm_type))

                objects_mdl = self.create_objects_model(elements=elements, elm_type=DeviceType(elm_type))

                return objects_mdl, elements, elm_type
            else:
                return None, None, None

        return None, None, None

    def view_objects_data(self):
        """
        On click, display the objects' properties
        """

        objects_mdl, elements, elm_type = self.get_selected_objects_model()

        if objects_mdl is not None:

            proxy = ObjectModelFilterProxy(mdl=objects_mdl)  # pass the same underlying list

            # update slice-view
            self.type_objects_list = elements
            self.ui.dataStructureTableView.setModel(proxy)

            # update time series view
            ts_mdl = gf.get_list_model(self.circuit.profile_magnitudes[elm_type][0])
            self.ui.device_type_magnitude_comboBox.setModel(ts_mdl)
            self.ui.device_type_magnitude_comboBox_2.setModel(ts_mdl)
            self.display_profiles(proxy_mdl=proxy)
            # the TS display will be triggered by the on-change event of the combobox

            # update the associations view
            assoc_mdl = gf.get_list_model(self.circuit.device_associations[elm_type])
            self.ui.associationsComboBox.setModel(assoc_mdl)
            self.display_associations(proxy_mdl=proxy)

        else:
            self.ui.dataStructureTableView.setModel(None)
            self.ui.device_type_magnitude_comboBox.clear()
            self.ui.device_type_magnitude_comboBox_2.clear()
            self.ui.associationsComboBox.clear()

    def get_selected_table_buses(self) -> Tuple[Set[dev.Bus], List[ALL_DEV_TYPES]]:
        """
        Get the list of selected buses, regardless of the object table type
        If the object has buses, this one takes them
        :return:
        """
        proxy_model: ObjectModelFilterProxy | None = self.get_current_objects_model_view()
        buses = set()
        selected_objects: List[ALL_DEV_TYPES] = list()

        if proxy_model is not None:

            source_model = proxy_model.sourceModel()

            sel_idx = self.ui.dataStructureTableView.selectedIndexes()
            objects: List[ALL_DEV_TYPES] = source_model.objects if hasattr(source_model, 'objects') else []

            if len(objects) > 0:

                if len(sel_idx) > 0:

                    unique = {proxy_model.mapToSource(idx).row() for idx in sel_idx}

                    for idx in unique:
                        sel_obj: ALL_DEV_TYPES = source_model.objects[idx]
                        selected_objects.append(sel_obj)

                    buses = self.circuit.get_buses_from_objects(elements=selected_objects,
                                                                dtype=objects[0].device_type)

        return buses, selected_objects

    def get_selected_table_substations(self) -> Tuple[Set[dev.Substation], List[ALL_DEV_TYPES]]:
        """
        Get the substations matching the table selection
        :return:  set of substations, list of selected objects originating the substation set
        """
        selected_objects = self.get_selected_db_table_objects()

        elm2se: Dict[ALL_DEV_TYPES, List[dev.Substation]] = dict()

        # Associate country, community, region and municipality to substation
        for se in self.circuit.substations:
            for elm in [se.country, se.community, se.region, se.municipality]:
                if elm is not None:
                    if elm in elm2se:
                        elm2se[elm].append(se)
                    else:
                        elm2se[elm] = [se]

        # associate voltage levels to substations
        for vl in self.circuit.voltage_levels:
            if vl.substation is not None:
                elm2se[vl] = [vl.substation]

        # associate buses to substations
        for bus in self.circuit.buses:
            if bus.substation is not None:
                elm2se[bus] = [bus.substation]

        substations = set()
        for sel_obj in selected_objects:
            se_list = elm2se.get(sel_obj, None)
            if se_list is not None:
                for se in se_list:
                    substations.add(se)

        return substations, selected_objects

    def delete_selected_db_table_objects(self):
        """
        Delete selection from the database main table
        """

        selected_objects = self.get_selected_db_table_objects()

        if len(selected_objects):

            ok = yes_no_question('Are you sure that you want to delete_with_dialogue the selected elements?', 'Delete')
            if ok:
                for obj in selected_objects:

                    # delete_with_dialogue from the database
                    self.circuit.delete_element(obj=obj)

                    # delete_with_dialogue from all diagrams
                    for diagram in self.diagram_widgets_list:
                        diagram.delete_element_utility_function(device=obj, propagate=False)

                # update the view
                self.view_objects_data()
                self.update_from_to_list_views()
                self.update_date_dependent_combos()

                self.show_info_toast(f"{len(selected_objects)} objects deleted")

    def duplicate_selected_db_table_objects(self):
        """
        Delete selection
        """

        selected_objects = self.get_selected_db_table_objects()

        if len(selected_objects):

            ok = yes_no_question('Are you sure that you want to duplicate the selected elements?',
                                 'Duplicate')
            if ok:
                for obj in selected_objects:
                    cpy = obj.copy(forced_new_idtag=True)
                    cpy.name += ' copy'
                    self.circuit.add_element(obj=cpy)

                # update the view
                self.view_objects_data()
                self.update_from_to_list_views()
                self.update_date_dependent_combos()

    def fuse_selected_db_table_objects(self):
        """
        Fuse selection
        """

        selected_objects = self.get_selected_db_table_objects()

        if len(selected_objects):

            if selected_objects[0].device_type == DeviceType.SubstationDevice:

                ok = yes_no_question('Are you sure that you want to merge the selected substations?',
                                     'Merge')
                if ok:
                    # merge substations into the first
                    self.circuit.merge_substations(selected_objects=selected_objects)

                    # update the view
                    self.view_objects_data()
                    self.update_from_to_list_views()
                    self.update_date_dependent_combos()

                    self.show_info_toast(f"{len(selected_objects)} substations merged")
            else:
                self.show_warning_toast(
                    f'Merge function not available for {selected_objects[0].device_type.value} devices')

    def copy_selected_idtag(self):
        """
        Copy selected idtags
        """

        selected_objects = self.get_selected_db_table_objects()

        if len(selected_objects):
            # copy to clipboard
            cb = QtWidgets.QApplication.clipboard()
            cb.clear()
            cb.setText("\n".join([obj.idtag for obj in selected_objects]))

            self.show_info_toast("Copied!")

    def add_objects_to_current_diagram(self):
        """
        Add selected DB objects to current diagram
        """

        selected_objects = self.get_selected_db_table_objects()

        if len(selected_objects):

            diagram = self.get_selected_diagram_widget()
            logger = bs.Logger()

            if isinstance(diagram, SchematicWidget):
                injections_by_bus = self.circuit.get_injection_devices_grouped_by_bus()
                injections_by_fluid_node = self.circuit.get_injection_devices_grouped_by_fluid_node()

                for device in selected_objects:
                    diagram.add_object_to_the_schematic(elm=device,
                                                        injections_by_bus=injections_by_bus,
                                                        injections_by_fluid_node=injections_by_fluid_node,
                                                        logger=logger)

            elif isinstance(diagram, GridMapWidget):

                for device in selected_objects:
                    diagram.add_object_to_the_schematic(elm=device, logger=logger)

            if len(logger):
                dlg = LogsDialogue(name="Add selected DB objects to current diagram", logger=logger)
                dlg.setModal(True)
                dlg.exec()

    def add_new_bus_diagram_from_selection(self):
        """
        Create a New diagram from a buses selection
        """
        selected_buses, selected_objects = self.get_selected_table_buses()

        if len(selected_buses):
            diagram = make_diagram_from_buses(circuit=self.circuit,
                                              buses=selected_buses,
                                              name=selected_objects[0].name + " diagram")

            diagram_widget = SchematicWidget(gui=self,
                                             circuit=self.circuit,
                                             diagram=diagram,
                                             default_bus_voltage=self.ui.defaultBusVoltageSpinBox.value(),
                                             time_index=self.get_diagram_slider_index())

            self.add_diagram_widget_and_diagram(diagram_widget=diagram_widget,
                                                diagram=diagram)
            self.set_diagrams_list_view()

            self.show_info_toast(f"{diagram.name} added")

    def add_new_map_from_database_selection(self):
        """
        Create a New map from a buses selection
        """

        # from whatever, get the selected substations
        selected_buses, selected_objects = self.get_selected_table_buses()

        if len(selected_buses):

            tpes = [
                DeviceType.SubstationDevice,
                DeviceType.LineDevice,
                DeviceType.DCLineDevice,
                DeviceType.HVDCLineDevice,
                DeviceType.GeneratorDevice,
                DeviceType.BatteryDevice,
                DeviceType.LoadDevice,
                DeviceType.StaticGeneratorDevice,
                DeviceType.ExternalGridDevice
            ]

            self.new_se_dlg = CheckListDialogue(
                objects_list=[e.value for e in tpes]
            )

            if self.circuit.get_substation_number() > 0:
                # showing this menu only makes sense if there is anything there
                self.new_se_dlg.exec()
            else:
                self.show_warning_toast("No substations to draw...")
                return

            cmap_text = self.ui.palette_comboBox.currentText()
            cmap = self.cmap_dict[cmap_text]
            subgrid = self.circuit.slice_buses(buses=list(selected_buses))

            diagram = generate_map_diagram(
                substations=subgrid.get_substations() if self.new_se_dlg.selected(
                    DeviceType.SubstationDevice.value) else list(),
                voltage_levels=subgrid.get_voltage_levels() if self.new_se_dlg.selected(
                    DeviceType.SubstationDevice.value) else list(),
                lines=subgrid.get_lines() if self.new_se_dlg.selected(DeviceType.LineDevice.value) else list(),
                dc_lines=subgrid.get_dc_lines() if self.new_se_dlg.selected(
                    DeviceType.DCLineDevice.value) else list(),
                hvdc_lines=subgrid.get_hvdc() if self.new_se_dlg.selected(
                    DeviceType.HVDCLineDevice.value) else list(),
                fluid_nodes=subgrid.get_fluid_nodes(),
                fluid_paths=subgrid.get_fluid_paths(),
                external_grids=subgrid.external_grids if self.new_se_dlg.selected(
                    DeviceType.ExternalGridDevice.value) else list(),
                static_generators=subgrid.static_generators if self.new_se_dlg.selected(
                    DeviceType.StaticGeneratorDevice.value) else list(),
                loads=subgrid.loads if self.new_se_dlg.selected(DeviceType.LoadDevice.value) else list(),
                batteries=subgrid.batteries if self.new_se_dlg.selected(
                    DeviceType.BatteryDevice.value) else list(),
                generators=subgrid.generators if self.new_se_dlg.selected(
                    DeviceType.GeneratorDevice.value) else list(),
                prog_func=None,
                text_func=None,
                name='Map diagram',
                use_flow_based_width=self.ui.branch_width_based_on_flow_checkBox.isChecked(),
                min_branch_width=self.ui.min_branch_size_spinBox.value(),
                max_branch_width=self.ui.max_branch_size_spinBox.value(),
                min_bus_width=self.ui.min_node_size_spinBox.value(),
                max_bus_width=self.ui.max_node_size_spinBox.value(),
                arrow_size=self.ui.arrow_size_size_spinBox.value(),
                palette=cmap,
                default_bus_voltage=self.ui.defaultBusVoltageSpinBox.value()
            )

            default_tile_source = self.tile_name_dict[self.ui.tile_provider_comboBox.currentText()]
            tile_source = self.tile_name_dict.get(diagram.tile_source, default_tile_source)

            diagram_widget = GridMapWidget(
                gui=self,
                tile_src=tile_source,
                start_level=diagram.start_level,
                longitude=diagram.longitude,
                latitude=diagram.latitude,
                name=diagram.name,
                circuit=self.circuit,
                diagram=diagram,
            )

            self.add_diagram_widget_and_diagram(diagram_widget=diagram_widget,
                                                diagram=diagram)
            self.set_diagrams_list_view()

            self.show_info_toast(f"{diagram.name} added")

    def crop_model_to_buses_selection(self):
        """
        Crop model to buses selection
        :return:
        """
        selected_buses, selected_objects = self.get_selected_table_buses()

        if len(selected_buses):

            ok = yes_no_question(
                text="This will delete all buses and their connected elements that were not selected."
                     "This cannot be undone and it is dangerous if you don't know"
                     "what you are doing. \nAre you sure?",
                title="Crop model to buses selection?")

            if ok:
                to_be_deleted = list()
                for bus in self.circuit.buses:
                    if bus not in selected_buses:
                        to_be_deleted.append(bus)

                for bus in to_be_deleted:
                    self.circuit.delete_bus(obj=bus, delete_associated=True)

                self.view_objects_data()  # re-paint the table

                self.show_info_toast(f"{len(to_be_deleted)} buses removed from the model")

    def grid_reduction_from_table_selection(self):
        """
        Crop model to buses selection
        """
        selected_buses, selected_objects = self.get_selected_table_buses()

        if len(selected_buses):
            # get the previous power flow
            _, pf_res = self.session.power_flow

            self.grid_reduction_dialogue = GridReduceDialogue(grid=self.circuit,
                                                              session=self.session,
                                                              selected_buses_set=selected_buses)

            self.grid_reduction_dialogue.exec()

            if self.grid_reduction_dialogue.did_reduce:

                # delete from the diagrams
                self.delete_from_all_diagrams(elements=list(selected_buses))

                # update the view
                self.view_objects_data()
                self.update_from_to_list_views()
                self.update_date_dependent_combos()
            else:
                self.show_warning_toast("No reduction done")
        else:
            self.show_warning_toast("Select some elements first")

    def grid_reduction_from_schematic_selection(self):
        """
        Call the grid reduction dialogue on the schematic selection
        """

        diagram_widget = self.get_selected_diagram_widget()
        if isinstance(diagram_widget, SchematicWidget):
            selected_buses = diagram_widget.get_selected_buses()
            selected_se = list()

        elif isinstance(diagram_widget, GridMapWidget):
            selected_se = diagram_widget.get_selected_substations()
            selected_buses = diagram_widget.get_selected_buses()
        else:
            return None

        if len(selected_buses):
            # get the previous power flow
            _, pf_res = self.session.power_flow

            selected_buses_set: Set[dev.Bus] = {bus for i, bus, graphic in selected_buses}

            self.grid_reduction_dialogue = GridReduceDialogue(grid=self.circuit,
                                                              session=self.session,
                                                              selected_buses_set=selected_buses_set)

            self.grid_reduction_dialogue.exec()

            if self.grid_reduction_dialogue.did_reduce:
                self.delete_from_all_diagrams(elements=[bus for i, bus, graphic in selected_buses])

            if isinstance(diagram_widget, GridMapWidget):
                # if this is a map, delete the elements from there
                for se_graphics in selected_se:
                    if se_graphics.api_object is not None:
                        diagram_widget.delete_element_utility_function(
                            device=se_graphics.api_object,
                            propagate=True
                        )

            self.show_info_toast("Done!")
        else:
            self.show_warning_toast("No selected buses :/")

    def add_objects(self):
        """
        Add default objects objects
        """
        model = self.get_current_objects_model_view()
        elm_type = self.get_db_object_selected_type()

        if model is not None and elm_type is not None:

            if elm_type == DeviceType.SubstationDevice.value:
                self.circuit.add_substation(dev.Substation(name=f'SE {self.circuit.get_substation_number() + 1}'))
                self.update_from_to_list_views()

            elif elm_type == DeviceType.VoltageLevelDevice.value:
                self.circuit.add_voltage_level(dev.VoltageLevel(
                    name=f'VL {self.circuit.get_voltage_levels_number() + 1}')
                )
                self.update_from_to_list_views()

            elif elm_type == DeviceType.BusBarDevice.value:
                self.circuit.add_bus_bar(dev.BusBar(name=f'BB {self.circuit.get_bus_bars_number() + 1}'))
                self.update_from_to_list_views()

            elif elm_type == DeviceType.ZoneDevice.value:
                self.circuit.add_zone(dev.Zone(name=f'Zone {self.circuit.get_zone_number() + 1}'))
                self.update_from_to_list_views()

            elif elm_type == DeviceType.AreaDevice.value:
                self.circuit.add_area(dev.Area(name=f'Area {self.circuit.get_area_number() + 1}'))
                self.update_from_to_list_views()

            elif elm_type == DeviceType.CountryDevice.value:
                self.circuit.add_country(dev.Country(name=f'Country {self.circuit.get_country_number() + 1}'))
                self.update_from_to_list_views()

            elif elm_type == DeviceType.CommunityDevice.value:
                self.circuit.add_community(dev.Community(
                    name=f'Community {self.circuit.get_communities_number() + 1}')
                )
                self.update_from_to_list_views()

            elif elm_type == DeviceType.RegionDevice.value:
                self.circuit.add_region(dev.Region(name=f'Region {self.circuit.get_regions_number() + 1}'))
                self.update_from_to_list_views()

            elif elm_type == DeviceType.MunicipalityDevice.value:
                self.circuit.add_municipality(dev.Municipality(
                    name=f'Municipalities {self.circuit.get_municipalities_number() + 1}')
                )
                self.update_from_to_list_views()

            elif elm_type == DeviceType.BusDevice.value:
                self.circuit.add_bus(dev.Bus(name=f'Bus {self.circuit.get_bus_number() + 1}'))

            elif elm_type == DeviceType.ContingencyGroupDevice.value:
                group = dev.ContingencyGroup(
                    name=f"Contingency group {self.circuit.get_contingency_groups_number() + 1}"
                )
                self.circuit.add_contingency_group(group)

            elif elm_type == DeviceType.RemedialActionGroupDevice.value:
                group = dev.RemedialActionGroup(
                    name=f"Remedial actions group {self.circuit.get_remedial_action_groups_number() + 1}"
                )
                self.circuit.add_remedial_action_group(group)

            elif elm_type == DeviceType.InvestmentsGroupDevice.value:
                group = dev.InvestmentsGroup(name=f"Investments group {len(self.circuit.investments_groups) + 1}")
                self.circuit.add_investments_group(group)

            elif elm_type == DeviceType.BranchGroupDevice.value:
                group = dev.BranchGroup(name=f"Branch group {self.circuit.get_branch_groups_number() + 1}")
                self.circuit.add_branch_group(group)

            elif elm_type == DeviceType.Technology.value:
                tech = dev.Technology(name=f"Technology {len(self.circuit.technologies) + 1}")
                self.circuit.add_technology(tech)

            elif elm_type == DeviceType.OverheadLineTypeDevice.value:

                obj = dev.OverheadLineType()
                obj.frequency = self.circuit.fBase
                obj.name = f'Tower {len(self.circuit.overhead_line_types) + 1}'
                self.circuit.add_overhead_line(obj)

            elif elm_type == DeviceType.UnderGroundLineDevice.value:

                name = f'Cable {len(self.circuit.underground_cable_types) + 1}'
                obj = dev.UndergroundLineType(name=name)
                self.circuit.add_underground_line(obj)

            elif elm_type == DeviceType.SequenceLineDevice.value:

                name = f'Sequence line {len(self.circuit.sequence_line_types) + 1}'
                obj = dev.SequenceLineType(name=name)
                self.circuit.add_sequence_line(obj)

            elif elm_type == DeviceType.WireDevice.value:

                name = f'Wire {len(self.circuit.wire_types) + 1}'
                obj = dev.Wire(name=name)
                self.circuit.add_wire(obj)

            elif elm_type == DeviceType.TransformerTypeDevice.value:

                name = f'Transformer type {len(self.circuit.transformer_types) + 1}'
                obj = dev.TransformerType(hv_nominal_voltage=10, lv_nominal_voltage=0.4, nominal_power=2,
                                          copper_losses=0.8, iron_losses=0.1, no_load_current=0.1,
                                          short_circuit_voltage=0.1,
                                          gr_hv1=0.5, gx_hv1=0.5, name=name)
                self.circuit.add_transformer_type(obj)

            elif elm_type == DeviceType.FuelDevice.value:

                name = f'Fuel {len(self.circuit.fuels) + 1}'
                obj = dev.Fuel(name=name)
                self.circuit.add_fuel(obj)

            elif elm_type == DeviceType.EmissionGasDevice.value:

                name = f'Gas {len(self.circuit.emission_gases) + 1}'
                obj = dev.EmissionGas(name=name)
                self.circuit.add_emission_gas(obj)

            elif elm_type == DeviceType.Owner.value:

                name = f'Owner {len(self.circuit.owners) + 1}'
                obj = dev.Owner(name=name)
                self.circuit.add_owner(obj)

            elif elm_type == DeviceType.ModellingAuthority.value:

                name = f'Modelling authority {self.circuit.get_modelling_authorities_number()}'
                obj = dev.ModellingAuthority(name=name)
                self.circuit.add_modelling_authority(obj)

            elif elm_type == DeviceType.FacilityDevice.value:

                name = f'Facility {self.circuit.get_facility_number()}'
                obj = dev.Facility(name=name)
                self.circuit.add_facility(obj)

            elif elm_type == DeviceType.DynamicModelHostDevice.value:

                name = f'RMS model {self.circuit.get_rms_models_number()}'
                obj = dev.DynamicModelHost(name=name)
                self.circuit.add_rms_model(obj)

            elif elm_type == DeviceType.RmsEventDevice.value:

                name = f'RMS event {len(self.circuit.rms_events)}'
                obj = dev.RmsEvent(name=name)
                self.circuit.add_rms_event(obj)

            elif elm_type == DeviceType.RmsEventsGroupDevice.value:

                name = f'RMS event group {len(self.circuit.rms_events_groups)}'
                obj = dev.RmsEventsGroup(name=name)
                self.circuit.add_rms_events_group(obj)

            else:
                info_msg("This object does not support table-like addition.\nUse the schematic instead.")
                return

            # update the view
            self.view_objects_data()

    def colour_branches_like_group(self):
        """
        Colour the branches like the branch group
        """
        model: ObjectModelFilterProxy | None = self.get_current_objects_model_view()

        if model is not None:

            if len(self.ui.dataStructuresTreeView.selectedIndexes()) > 0:

                rows = np.unique([sel.row() for sel in self.ui.dataStructureTableView.selectedIndexes()])

                for i in rows:

                    elm: dev.BranchGroup = model.objects[i]

                    if elm.device_type == DeviceType.BranchGroupDevice:

                        for br in self.circuit.get_branches_iter(add_vsc=True,
                                                                 add_hvdc=True,
                                                                 add_switch=True):
                            if br.group == elm:
                                br.color = elm.color

                self.show_info_toast("Done!")
            else:
                self.show_warning_toast("Nothing selected :/")

    def launch_object_editor(self):
        """
        Edit catalogue element
        """
        model: ObjectModelFilterProxy | None = self.get_current_objects_model_view()

        if model is not None:

            if len(self.ui.dataStructuresTreeView.selectedIndexes()) > 0:

                sel_item = self.ui.dataStructuresTreeView.selectedIndexes()[0]
                elm_type = sel_item.data(role=QtCore.Qt.ItemDataRole.DisplayRole)

                # get the selected index
                idx_proxy = self.ui.dataStructureTableView.currentIndex().row()

                if idx_proxy > -1:

                    # get the object from the table itself
                    elm = model.objects[idx_proxy]

                    if elm_type == DeviceType.OverheadLineTypeDevice.value:

                        # launch editor
                        self.tower_builder_window = TowerBuilderGUI(
                            tower=elm,
                            wires_catalogue=self.circuit.wire_types
                        )
                        self.tower_builder_window.setModal(True)
                        self.tower_builder_window.resize(int(1.81 * 700.0), 700)
                        self.tower_builder_window.exec()

                    elif elm_type == DeviceType.DynamicModelHostDevice.value:

                        self.rms_model_Editor_window = RmsModelEditorGUI(self.circuit.rms_models, model_host=elm)
                        self.rms_model_Editor_window.resize(int(1.81 * 700.0), 700)
                        self.rms_model_Editor_window.show()

                    elif elm_type == DeviceType.LineDevice.value:
                        dlg = LineEditor(line=elm, grid=self.circuit)
                        dlg.exec()

                    elif elm_type == DeviceType.Transformer2WDevice.value:
                        dlg = TransformerEditor(branch=elm, grid=self.circuit, modify_on_accept=True)
                        if dlg.exec():
                            pass

                    elif elm_type == DeviceType.ControllableShuntDevice.value:
                        dlg = ControllableShuntEditor(api_object=elm)
                        if dlg.exec():
                            pass

                    elif elm_type == DeviceType.Transformer3WDevice.value:
                        Sbase = self.circuit.Sbase
                        dlg = Transformer3WEditor(elm, Sbase, modify_on_accept=True)
                        if dlg.exec():
                            pass

                    elif elm_type == DeviceType.RmsEventDevice.value:
                        lines = []
                        injections = []
                        for line in self.circuit.get_branches_iter(add_vsc=True, add_hvdc=True, add_switch=True):
                            lines.append(line)
                        for injection in self.circuit.get_injection_devices_iter():
                            injections.append(injection)

                        all_devices = lines + injections

                        self.rms_events_Editor_window = RmsEventEditor(devices=all_devices,
                                                                       event=elm)
                        self.rms_events_Editor_window.resize(int(1.81 * 700.0), 700)

                        # TODO: check if we are actually editing the object directly
                        # if self.rms_events_Editor_window.exec() == QtWidgets.QDialog.DialogCode.Accepted:
                        #     # update event
                        #     self.circuit.rms_events[idx] = self.rms_events_Editor_window.get_updated_event()

                    else:

                        warning_msg('No editor available.\n'
                                    'The values can be changed from the table or '
                                    'via context menus in the graphical interface.',
                                    'Edit')
                else:
                    info_msg('Choose an element from the table')
            else:
                info_msg('Choose an element from the table')
        else:
            info_msg('Select a catalogue element and then a catalogue object')

    def set_value_to_column(self):
        """
        Set the value to all the column
        :return: Nothing
        """
        idx = self.ui.dataStructureTableView.currentIndex()
        mdl: ObjectModelFilterProxy | None = self.get_current_objects_model_view()
        col = idx.column()
        if mdl is not None:
            if col > -1:
                mdl.copy_to_column(idx)
                # update the view
                self.view_objects_data()
            else:
                info_msg('Select some element to serve as source to copy', 'Set value to column')
        else:
            pass

    def highlight_selection_buses(self):
        """
        Highlight and select the buses of the selected objects
        """

        model: ObjectModelFilterProxy | None = self.get_current_objects_model_view()

        if model is not None:

            sel_idx = self.ui.dataStructureTableView.selectedIndexes()
            objects = model.objects

            if len(objects) > 0:

                if len(sel_idx) > 0:

                    unique = set()
                    for idx in sel_idx:
                        unique.add(idx.row())
                    sel_obj = list()
                    for idx in unique:
                        sel_obj.append(objects[idx])

                    elm = objects[0]

                    self.clear_big_bus_markers()
                    color = QtGui.QColor(55, 200, 171, 180)

                    if elm.device_type == DeviceType.BusDevice:

                        self.set_big_bus_marker(buses=sel_obj, color=color)

                    elif elm.device_type in [DeviceType.BranchDevice,
                                             DeviceType.LineDevice,
                                             DeviceType.Transformer2WDevice,
                                             DeviceType.HVDCLineDevice,
                                             DeviceType.VscDevice,
                                             DeviceType.DCLineDevice]:
                        buses = list()
                        for br in sel_obj:
                            buses.append(br.bus_from)
                            buses.append(br.bus_to)
                        self.set_big_bus_marker(buses=buses, color=color)

                    else:
                        buses = list()
                        for elm in sel_obj:
                            buses.append(elm.bus)
                        self.set_big_bus_marker(buses=buses, color=color)

                else:
                    info_msg('Select some elements to highlight', 'Highlight')
            else:
                pass

    def get_objects_time_index(self) -> Union[None, int]:
        """
        Get the time index of the objects slider already
        accounting for the -1 -> None conversion
        :return: None or int
        """
        t_idx = self.ui.db_step_slider.value()
        if t_idx <= -1:
            return None
        else:
            return t_idx

    def highlight_based_on_property(self):
        """
        Highlight and select the buses of the selected objects
        """
        indices = self.ui.dataStructureTableView.selectedIndexes()

        if len(indices):
            model: ObjectModelFilterProxy | None = self.get_current_objects_model_view()

            if model is not None:
                objects = model.objects
                t_idx = self.get_objects_time_index()

                if len(objects) > 0:
                    col_indices = list({index.column() for index in indices})
                    elm = objects[0]
                    attr = model.attributes[col_indices[0]]
                    gc_prop = elm.registered_properties[attr]
                    if gc_prop is None:
                        info_msg(f"The proprty {attr} cannot be found :(", "Highlight based on property")
                        return

                    if gc_prop.tpe in [float, int]:

                        self.clear_big_bus_markers()

                        if elm.device_type == DeviceType.BusDevice:
                            # buses
                            buses = objects
                            values = [elm.get_value(prop=gc_prop, t_idx=t_idx) for elm in objects]

                        elif elm.device_type in [DeviceType.BranchDevice,
                                                 DeviceType.LineDevice,
                                                 DeviceType.DCLineDevice,
                                                 DeviceType.HVDCLineDevice,
                                                 DeviceType.Transformer2WDevice,
                                                 DeviceType.SwitchDevice,
                                                 DeviceType.VscDevice,
                                                 DeviceType.UpfcDevice]:
                            # Branches
                            buses = list()
                            values = list()
                            for br in objects:
                                gc_prop = br.registered_properties[attr]
                                buses.append(br.bus_from)
                                buses.append(br.bus_to)
                                val = elm.get_value(prop=gc_prop, t_idx=t_idx)
                                values.append(val)
                                values.append(val)

                        else:
                            # loads, generators, etc...
                            buses = list()
                            values = list()
                            for elm in objects:
                                gc_prop = elm.registered_properties[attr]
                                val = elm.get_value(prop=gc_prop, t_idx=t_idx)
                                buses.append(elm.bus)
                                values.append(val)

                        # build the color map
                        seq = [(0.0, 'gray'),
                               (0.5, 'orange'),
                               (1, 'red')]
                        cmap = LinearSegmentedColormap.from_list('lcolors', seq)
                        mx = max(values)

                        if mx != 0:

                            colors = np.zeros(len(values), dtype=object)
                            for i, value in enumerate(values):
                                r, g, b, a = cmap(value / mx)
                                colors[i] = QtGui.QColor(r * 255, g * 255, b * 255, a * 255)

                            # color based on the value
                            self.set_big_bus_marker_colours(buses=buses, colors=colors, tool_tips=None)

                        else:
                            info_msg('The maximum value is 0, so the coloring cannot be applied',
                                     'Highlight based on property')
                    else:
                        info_msg('The selected property must be of a numeric type',
                                 'Highlight based on property')

                else:
                    pass

    def assign_to_profile(self):
        """
        Assign the snapshot values at the object DB to the profiles
        """
        indices = self.ui.dataStructureTableView.selectedIndexes()

        if len(indices):
            model: ObjectModelFilterProxy | None = self.get_current_objects_model_view()

            if model is not None:
                logger = bs.Logger()

                t_idx = self.get_objects_time_index()
                attr_list = list()
                for index in indices:
                    i = index.row()
                    p_idx = index.column()
                    elm = model.objects[i]
                    attr = model.attributes[p_idx]
                    gc_prop = elm.registered_properties[attr]
                    attr_list.append(attr)
                    if gc_prop.has_profile():
                        val = elm.get_value(prop=gc_prop, t_idx=t_idx)
                        profile = elm.get_profile_by_prop(prop=gc_prop)
                        profile.fill(val)
                    else:
                        logger.add_error("No profile found for " + attr, device=elm.name)

                if logger.size():
                    logs_window = LogsDialogue("Assign to profile", logger=logger)
                    logs_window.exec()
                else:
                    lst = ", ".join(attr_list)
                    self.show_info_toast(f"{lst} assigned to profile")
        else:
            info_msg("Select a cell or a column first", "Assign to profile")

    def objects_histogram_analysis_plot(self):
        """
        Histogram analysis
        :return:
        """
        if len(self.ui.dataStructuresTreeView.selectedIndexes()) > 0:
            elm_type = self.ui.dataStructuresTreeView.selectedIndexes()[0].data(role=QtCore.Qt.ItemDataRole.DisplayRole)

            if len(self.circuit.get_elements_by_type(device_type=DeviceType(elm_type))):
                object_histogram_analysis(circuit=self.circuit,
                                          object_type=elm_type,
                                          t_idx=self.get_db_slider_index(),
                                          fig=None)
                plt.show()
        else:
            info_msg('Select a data structure')

    def objects_smart_search(self):
        """
        Objects and time series object-based filtering
        :return:
        """

        # gather the model, which for sure is a ExpressionFilterProxy or None
        proxy_mdl: ObjectModelFilterProxy | None = self.get_current_objects_model_view()

        if proxy_mdl is not None:
            if len(proxy_mdl.all_objects) > 0:

                has_err, err_txt = proxy_mdl.setExpression(self.ui.smart_search_lineEdit.text())

                # display time series
                self.display_profiles(proxy_mdl=proxy_mdl)

                # display associations
                self.display_associations(proxy_mdl=proxy_mdl)

                if has_err:
                    self.show_error_toast(err_txt)

            else:
                # nothing to search
                self.show_warning_toast("The collection is empty...")

        else:
            self.show_warning_toast("Nothing to search on, select an object...")

        return None

    def delete_inconsistencies(self):
        """
        Call delete_with_dialogue shit
        :return:
        """
        ok = yes_no_question(
            "This action removes all disconnected devices with no active profile and delete all small islands",
            "Delete inconsistencies")

        if ok:
            logger = self.delete_shit()

            if len(logger) > 0:
                dlg = LogsDialogue("Delete inconsistencies", logger)
                dlg.setModal(True)
                dlg.exec()

    def delete_shit(self, min_island=1):
        """
        Delete small islands, disconnected stuff and other garbage
        """
        numerical_circuit_ = compile_numerical_circuit_at(circuit=self.circuit, )
        islands = numerical_circuit_.split_into_islands()
        logger = bs.Logger()
        buses_to_delete = list()
        buses_to_delete_idx = list()
        for island in islands:
            if island.nbus <= min_island:
                for r in island.original_bus_idx:
                    buses_to_delete.append(self.circuit.buses[r])
                    buses_to_delete_idx.append(r)

        for r, bus in enumerate(self.circuit.buses):
            if not bus.active and not np.any(bus.active_prof.toarray()):
                if r not in buses_to_delete_idx:
                    buses_to_delete.append(bus)
                    buses_to_delete_idx.append(r)

        # delete_with_dialogue the grphics from all diagrams
        self.delete_from_all_diagrams(elements=buses_to_delete)

        for elm in buses_to_delete:
            logger.add_info("Deleted " + str(elm.device_type.value), elm.name)

        # search other elements to delete_with_dialogue
        for dev_lst in [self.circuit.lines,
                        self.circuit.dc_lines,
                        self.circuit.vsc_devices,
                        self.circuit.hvdc_lines,
                        self.circuit.transformers2w,
                        self.circuit.get_generators(),
                        self.circuit.get_loads(),
                        self.circuit.get_shunts(),
                        self.circuit.get_batteries(),
                        self.circuit.get_static_generators()]:

            for elm in dev_lst:
                if not elm.active and not np.any(elm.active_prof.toarray()):
                    self.delete_from_all_diagrams(elements=[elm])
                    logger.add_info("Deleted " + str(elm.device_type.value), elm.name)

        return logger

    def clean_database(self):
        """
        Clean the DataBase
        """

        ok = yes_no_question("This action may delete_with_dialogue unused objects and references, \nAre you sure?",
                             title="DB clean")

        if ok:
            logger = self.circuit.clean()

            if len(logger) > 0:
                dlg = LogsDialogue('DB clean logger', logger)
                dlg.exec()

    def scale(self):
        """
        Show the system scaler window
        The scaler window may modify the circuit
        """
        system_scaler_window = SystemScaler(grid=self.circuit, parent=self)
        system_scaler_window.exec()

    def detect_substations(self):
        """
        Call the detect substations logic
        """

        ok = yes_no_question("Do you want to try to detect substations and voltage levels in the grid model?",
                             "Detect substations")

        if ok:
            val = 1.0 / (10.0 ** self.ui.rxThresholdSpinBox.value())
            detect_substations(grid=self.circuit,
                               r_x_threshold=val)

    def detect_facilities(self):
        """
        Call the detect facilities logic
        """
        ok = yes_no_question("Do you want to try to detect facilities in the grid model?",
                             "Detect facilities")

        if ok:
            detect_facilities(grid=self.circuit)

    def show_objects_context_menu(self, pos: QtCore.QPoint):
        """
        Show diagrams list view context menu
        :param pos: Relative click position
        """
        if len(self.ui.dataStructuresTreeView.selectedIndexes()) > 0:
            sel_item = self.ui.dataStructuresTreeView.selectedIndexes()[0]
            elm_type = sel_item.data(role=QtCore.Qt.ItemDataRole.DisplayRole)

            context_menu = QtWidgets.QMenu(parent=self.ui.diagramsListView)

            gf.add_menu_entry(menu=context_menu,
                              text="Edit",
                              icon_path=":/Icons/icons/edit.png",
                              function_ptr=self.launch_object_editor)

            gf.add_menu_entry(menu=context_menu,
                              text="Add",
                              icon_path=":/Icons/icons/plus.png",
                              function_ptr=self.add_objects)

            gf.add_menu_entry(menu=context_menu,
                              text="Delete",
                              icon_path=":/Icons/icons/minus.png",
                              function_ptr=self.delete_selected_db_table_objects)

            gf.add_menu_entry(menu=context_menu,
                              text="Duplicate object",
                              icon_path=":/Icons/icons/copy.png",
                              function_ptr=self.duplicate_selected_db_table_objects)

            gf.add_menu_entry(menu=context_menu,
                              text="Merge",
                              icon_path=":/Icons/icons/fusion.png",
                              function_ptr=self.fuse_selected_db_table_objects)

            gf.add_menu_entry(menu=context_menu,
                              text="Copy idtag",
                              icon_path=":/Icons/icons/copy.png",
                              function_ptr=self.copy_selected_idtag)

            gf.add_menu_entry(menu=context_menu,
                              text="Crop model to buses selection",
                              icon_path=":/Icons/icons/schematic.png",
                              function_ptr=self.crop_model_to_buses_selection)

            gf.add_menu_entry(menu=context_menu,
                              text="Grid reduction",
                              icon_path=":/Icons/icons/schematic.png",
                              function_ptr=self.grid_reduction_from_table_selection)

            gf.add_menu_entry(menu=context_menu,
                              text="Copy table",
                              icon_path=":/Icons/icons/copy.png",
                              function_ptr=self.copy_objects_data)

            gf.add_menu_entry(menu=context_menu,
                              text="Set value to column",
                              icon_path=":/Icons/icons/copy2down.png",
                              function_ptr=self.set_value_to_column)

            gf.add_menu_entry(menu=context_menu,
                              text="Assign to profile",
                              icon_path=":/Icons/icons/assign_to_profile.png",
                              function_ptr=self.assign_to_profile)

            if elm_type == DeviceType.BranchGroupDevice.value:
                gf.add_menu_entry(menu=context_menu,
                                  text="Colour branches like this",
                                  icon_path=":/Icons/icons/assign_to_profile.png",
                                  function_ptr=self.colour_branches_like_group)

            context_menu.addSeparator()

            gf.add_menu_entry(menu=context_menu,
                              text="New vicinity diagram",
                              icon_path=":/Icons/icons/grid_icon.png",
                              function_ptr=self.add_bus_vicinity_diagram_from_model)

            gf.add_menu_entry(menu=context_menu,
                              text="New diagram from selection",
                              icon_path=":/Icons/icons/schematicadd_to.png",
                              function_ptr=self.add_new_bus_diagram_from_selection)

            gf.add_menu_entry(menu=context_menu,
                              text="Add to current diagram",
                              icon_path=":/Icons/icons/schematicadd_to.png",
                              function_ptr=self.add_objects_to_current_diagram)

            gf.add_menu_entry(menu=context_menu,
                              text="Highlight buses selection",
                              icon_path=":/Icons/icons/highlight.png",
                              function_ptr=self.highlight_selection_buses)

            gf.add_menu_entry(menu=context_menu,
                              text="Highlight based on property",
                              icon_path=":/Icons/icons/highlight2.png",
                              function_ptr=self.highlight_based_on_property)

            context_menu.addSeparator()

            gf.add_menu_entry(menu=context_menu,
                              text="New map from selection",
                              icon_path=":/Icons/icons/map.png",
                              function_ptr=self.add_new_map_from_database_selection)

            # Convert global position to local position of the list widget
            mapped_pos = self.ui.dataStructureTableView.viewport().mapToGlobal(pos)
            context_menu.exec(mapped_pos)

        else:
            pass

    def add_substation_with_wizard(self):
        """
        Add substation with all its objects using a wizard
        :return:
        """

        # Get selected buses from the current diagram if any
        selected_buses_tuples = self.get_diagram_selected_buses()
        buses_to_replace = [bus for _, bus, _ in selected_buses_tuples] if selected_buses_tuples else None

        kv = self.get_default_voltage()
        dlg = SubstationDesigner(grid=self.circuit, default_voltage=kv)
        dlg.exec()
        if dlg.was_ok():

            se_object, voltage_levels = substation_wizards.create_substation(
                grid=self.circuit,
                se_name=dlg.get_name(),
                se_code=dlg.get_code(),
                lat=dlg.get_latitude(),
                lon=dlg.get_longitude(),
                vl_templates=dlg.get_voltage_levels(),
                buses_to_replace=buses_to_replace
            )

            dlg3 = CustomQuestionDialogue(title="New substation",
                                          question="How do you want to represent the merged grid?",
                                          answer1="Create new diagram",
                                          answer2="Add to current diagram")
            dlg3.exec()

            if dlg3.accepted_answer == 1:
                # Create a blank diagram and add to it
                self.new_bus_branch_diagram_from_substation(substations=[se_object])

            elif dlg3.accepted_answer == 2:
                self.add_substation_to_current_diagram(substations=[se_object])

            else:
                # not imported
                return

    def search_ts_point(self):
        """

        :return:
        """

        txt = self.ui.goToTsPointComboBox.currentText()
        mode: TimeSeriesSearchPoint = self.ts_search_points_dict[txt]

        if mode == TimeSeriesSearchPoint.LowestLoad:

            total = 0
            for elm in self.circuit.loads:
                total += elm.P_prof.toarray()

            idx = np.argmin(total)
            self.ui.db_step_slider.setValue(idx)

        elif mode == TimeSeriesSearchPoint.HighestLoad:

            total = 0
            for elm in self.circuit.loads:
                total += elm.P_prof.toarray()

            idx = np.argmax(total)
            self.ui.db_step_slider.setValue(idx)

        else:
            return





# MIT License
#
# Copyright (c) 2018 Ross Wilson
# Copyright (c) 2024, Santiago Peñate Vera
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
A server Tiles object for pySlipQt tiles.

All server tile sources should inherit from this class.
For example, see osm_tiles.py.
"""
import queue
import ssl
from urllib.request import Request, urlopen
from collections.abc import Callable
from PySide6.QtCore import QThread
from PySide6.QtGui import QPixmap

# SSL magic to solve the certificates hell
# https://stackoverflow.com/questions/68275857/urllib-error-urlerror-urlopen-error-ssl-certificate-verify-failed-certifica
ssl._create_default_https_context = ssl._create_stdlib_context


def log(val: str):
    print(val)


class TileWorker(QThread):
    """Thread class that gets request from queue, loads tile, calls callback."""

    def __init__(self,
                 id_num: int,
                 server: str,
                 tile_path: str,
                 requests_cue: queue.Queue,
                 callback: Callable[[int, float, float, QPixmap, bool], None],  # level, x, y, pixmap, error
                 error_tile: QPixmap,
                 content_type: str,
                 re_request_age: float,
                 error_image: QPixmap,
                 refresh_tiles_after_days=60):
        """
        Prepare the tile worker
        Results are returned in the callback() params.
        :param id_num: a unique numer identifying the worker instance
        :param server: server URL
        :param tile_path: path to tile on server
        :param requests_cue: the request queue
        :param callback: function to call after tile available
        :param error_tile: image of error tile
        :param content_type: expected Content-Type string
        :param re_request_age: number of days in tile age before re-requesting (0 means don't update tiles)
        :param error_image: the image to return on some error
        :param refresh_tiles_after_days:
        """
        super().__init__()

        self.id_num = id_num
        self.server = server
        self.tile_path = tile_path
        self.requests_cue = requests_cue
        self.callback: Callable[[int, float, float, QPixmap, bool], None] = callback
        self.error_tile_image = error_tile
        self.content_type = content_type
        self.re_request_age = re_request_age
        self.error_image = error_image
        self.daemon = True
        self.refresh_tiles_after_days = refresh_tiles_after_days

    def run(self):
        """

        :return:
        """
        while True:
            # get zoom level and tile coordinates to retrieve
            (level, x, y) = self.requests_cue.get()

            # try to retrieve the image
            error = False
            pixmap = self.error_image
            tile_url = self.server + self.tile_path.format(Z=level, X=x, Y=y)
            try:

                # Create a Request object with the desired headers
                response = urlopen(Request(tile_url, headers={'User-Agent': 'VeraGrid 5'}))

                content_type = response.info().get_content_type()

                if content_type == self.content_type:
                    data = response.read()
                    pixmap = QPixmap()
                    pixmap.loadFromData(data)
                else:
                    # show error, don't cache returned error tile
                    error = True
            except Exception as e:
                error = True
                log(f"{e} exception getting tile ({level},{x},{y}) with {tile_url}")

            # call the callback function passing level, x, y and pixmap data
            # error is False if we want to cache this tile on-disk
            self.callback(level, x, y, pixmap, error)

            # finally, removes request from queue
            self.requests_cue.task_done()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.  
# SPDX-License-Identifier: MPL-2.0

from __future__ import annotations
from typing import TYPE_CHECKING, Union
from PySide6.QtWidgets import QMenu
from VeraGrid.Gui.gui_functions import add_menu_entry
from VeraGrid.Gui.Diagrams.SchematicWidget.Branches.line_graphics_template import LineGraphicTemplateItem
from VeraGrid.Gui.Diagrams.SchematicWidget.terminal_item import BarTerminalItem, RoundTerminalItem
from VeraGrid.Gui.messages import yes_no_question
from VeraGrid.Gui.Diagrams.Editors.transformer_editor import TransformerEditor
from VeraGrid.Gui.Diagrams.Editors.transformer_taps_editor import TransformerTapsEditor
from VeraGridEngine.Devices.Branches.transformer import Transformer2W, TransformerType
from VeraGridEngine.enumerations import DeviceType, TapModuleControl

if TYPE_CHECKING:  # Only imports the below statements during type checking
    from VeraGrid.Gui.Diagrams.SchematicWidget.schematic_widget import SchematicWidget


class TransformerGraphicItem(LineGraphicTemplateItem):
    """
    TransformerGraphicItem
    """

    def __init__(self,
                 from_port: Union[BarTerminalItem, RoundTerminalItem],
                 to_port: Union[BarTerminalItem, RoundTerminalItem],
                 editor: SchematicWidget,
                 width=5,
                 api_object: Transformer2W = None,
                 draw_labels: bool = True):
        """

        :param from_port:
        :param to_port:
        :param editor:
        :param width:
        :param api_object:
        """
        LineGraphicTemplateItem.__init__(self=self,
                                         from_port=from_port,
                                         to_port=to_port,
                                         editor=editor,
                                         width=width,
                                         api_object=api_object,
                                         draw_labels=draw_labels)

    @property
    def api_object(self) -> Transformer2W:
        return self._api_object

    def contextMenuEvent(self, event):
        """
        Show context menu
        @param event:
        @return:
        """
        if self.api_object is not None:
            menu = QMenu()
            menu.addSection("Transformer")

            add_menu_entry(menu=menu,
                           text="Active",
                           function_ptr=self.enable_disable_toggle,
                           checkeable=True,
                           checked_value=self.api_object.active)

            add_menu_entry(menu=menu,
                           text="Draw labels",
                           function_ptr=self.enable_disable_label_drawing,
                           checkeable=True,
                           checked_value=self.draw_labels)

            add_menu_entry(menu=menu,
                           text="Delete",
                           function_ptr=self.delete,
                           icon_path=":/Icons/icons/delete3.png")

            add_menu_entry(menu=menu,
                           text="Edit template",
                           function_ptr=self.edit,
                           icon_path=":/Icons/icons/edit.png")

            menu.addSection('Tap changer')

            add_menu_entry(menu=menu,
                           text="Edit tap changer",
                           function_ptr=self.edit_tap_changer,
                           icon_path=":/Icons/icons/edit.png")

            add_menu_entry(menu=menu,
                           text="Tap up",
                           function_ptr=self.tap_up,
                           icon_path=":/Icons/icons/up.png")

            add_menu_entry(menu=menu,
                           text="Tap down",
                           function_ptr=self.tap_down,
                           icon_path=":/Icons/icons/down.png")

            add_menu_entry(menu=menu,
                           text="Control V from",
                           function_ptr=self.control_v_from,
                           icon_path=":/Icons/icons/edit.png")

            add_menu_entry(menu=menu,
                           text="Control V to",
                           function_ptr=self.control_v_to,
                           icon_path=":/Icons/icons/edit.png")

            menu.addSeparator()

            add_menu_entry(menu=menu,
                           text="Plot profiles",
                           function_ptr=self.plot_profiles,
                           icon_path=":/Icons/icons/plot.png")

            add_menu_entry(menu=menu,
                           text="Add to catalogue",
                           function_ptr=self.add_to_catalogue,
                           icon_path=":/Icons/icons/Catalogue.png")

            add_menu_entry(menu=menu,
                           text="Assign rate to profile",
                           function_ptr=self.assign_rate_to_profile,
                           icon_path=":/Icons/icons/assign_to_profile.png")

            add_menu_entry(menu=menu,
                           text="Assign active state to profile",
                           function_ptr=self.assign_status_to_profile,
                           icon_path=":/Icons/icons/assign_to_profile.png")

            add_menu_entry(menu=menu,
                           text="Flip",
                           function_ptr=self.flip_connections,
                           icon_path=":/Icons/icons/redo.png")

            add_menu_entry(menu=menu,
                           text="Change bus",
                           function_ptr=self.change_bus,
                           icon_path=":/Icons/icons/move_bus.png")

            menu.exec_(event.screenPos())

        else:
            pass

    def mouseDoubleClickEvent(self, event):
        """
        On double click, edit
        :param event:
        :return:
        """

        if self.api_object is not None:
            if self.api_object.device_type in [DeviceType.Transformer2WDevice, DeviceType.LineDevice]:
                # trigger the editor
                self.edit()
            elif self.api_object.device_type is DeviceType.SwitchDevice:
                # change state
                self.enable_disable_toggle()

    def edit(self):
        """
        Open the appropriate editor dialogue
        :return:
        """
        dlg = TransformerEditor(self.api_object, grid=self.editor.circuit, modify_on_accept=True)
        if dlg.exec():
            pass

    def edit_tap_changer(self):
        """

        :return:
        """

        dlg = TransformerTapsEditor(api_object=self.api_object.tap_changer)
        if dlg.exec():
            pass

    def show_transformer_editor(self):
        """
        Open the appropriate editor dialogue
        :return:
        """
        Sbase = self.editor.circuit.Sbase

        if self.api_object.template is not None:
            # automatically pick the template
            if isinstance(self.api_object.template, TransformerType):
                self.editor.circuit.add_transformer_type(self.api_object.template)
            else:
                # raise dialogue to set the template
                dlg = TransformerEditor(self.api_object, grid=self.editor.circuit, modify_on_accept=False)
                if dlg.exec():
                    tpe = dlg.get_template()
                    self.editor.circuit.add_transformer_type(tpe)
        else:
            # raise dialogue to set the template
            dlg = TransformerEditor(self.api_object, grid=self.editor.circuit, modify_on_accept=False)
            if dlg.exec():
                tpe = dlg.get_template()
                self.editor.circuit.add_transformer_type(tpe)

    def add_to_catalogue(self):
        """
        Add this object to the catalogue
        :return:
        """

        ok = yes_no_question(text="A template will be generated using this transformer values",
                             title="Add transformer type")

        if ok:
            tpe = self.api_object.get_transformer_type(Sbase=self.editor.circuit.Sbase)

            self.editor.circuit.add_transformer_type(tpe)

    def flip_connections(self):
        """
        Flip connections
        :return:
        """
        self.api_object.flip()

    def tap_up(self):
        """
        Set one tap up
        """
        self.api_object.tap_up()

    def tap_down(self):
        """
        Set one tap down
        """
        self.api_object.tap_down()

    def control_v_from(self):
        """

        :return:
        """
        self.api_object.regulation_bus = self.api_object.bus_from
        self.api_object.tap_module_control_mode = TapModuleControl.Vm

    def control_v_to(self):
        """

        :return:
        """
        self.api_object.regulation_bus = self.api_object.bus_to
        self.api_object.tap_module_control_mode = TapModuleControl.Vm



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
from __future__ import annotations

from typing import TYPE_CHECKING
import os
from PySide6 import QtWidgets
from VeraGrid.Gui.CGMESDialogue.cgmes_export_gui import Ui_CgmesExportDialog
import VeraGrid.Session.file_handler as filedrv
import VeraGrid.Gui.gui_functions as gf
from VeraGridEngine.enumerations import CGMESVersions
from VeraGridEngine.IO.cim.cgmes.cgmes_export import get_available_cgmes_profiles
from VeraGridEngine.IO.cim.cgmes.cgmes_enums import CgmesProfileType

if TYPE_CHECKING:
    from VeraGrid.Gui.Main.SubClasses.io import IoMain


class CgmesExportDialogue(QtWidgets.QDialog):
    """
    GridMergeDialogue
    """

    def __init__(self, app: IoMain):
        """
        GridMergeDialogue
        :param app: VeraGridMainGUI
        """
        QtWidgets.QDialog.__init__(self)
        self.ui = Ui_CgmesExportDialog()
        self.ui.setupUi(self)
        self.setWindowTitle('CGMES export')
        self.setModal(True)

        self.grid = app.circuit

        self.sessions_data = app.session

        self.project_directory = app.project_directory

        self.app: IoMain = app

        self.current_boundary_set = ""

        self.cgmes_version_dict = {x.value: x for x in [CGMESVersions.v2_4_15,
                                                        CGMESVersions.v3_0_0]}
        self.ui.cgmes_version_comboBox.setModel(gf.get_list_model(list(self.cgmes_version_dict.keys())))

        self.cgmes_profiles_dict = {key: CgmesProfileType(key) for key, val in
                                    get_available_cgmes_profiles(cgmes_version=CGMESVersions.v2_4_15).items()}

        self.ui.cgmes_profiles_listView.setModel(gf.get_list_model(list(self.cgmes_profiles_dict.keys()),
                                                                   checks=True, check_value=True))

        # Buttons
        self.ui.selectCGMESBoundarySetButton.clicked.connect(self.select_cgmes_boundary_set)
        self.ui.exportButton.clicked.connect(self.export)

        # change
        self.ui.cgmes_version_comboBox.currentTextChanged.connect(self.cgmes_version_change)

    def cgmes_version_change(self):
        """
        On GUI cgmes version change, display only the supported profiles
        """
        cgmes_version = self.cgmes_version_dict[self.ui.cgmes_version_comboBox.currentText()]

        self.cgmes_profiles_dict = {key: CgmesProfileType(key) for key, val in
                                    get_available_cgmes_profiles(cgmes_version=cgmes_version).items()}

        self.ui.cgmes_profiles_listView.setModel(gf.get_list_model(list(self.cgmes_profiles_dict.keys()),
                                                                   checks=True, check_value=True))

    def select_cgmes_boundary_set(self):
        """
        Select the current boundary set
        """
        files_types = "Boundary set (*.zip)"

        dialogue = QtWidgets.QFileDialog(None,
                                         caption='Select Boundary set file',
                                         directory=self.project_directory,
                                         filter=files_types)
        if dialogue.exec():
            filenames = dialogue.selectedFiles()
            if len(filenames) > 0:
                self.current_boundary_set = filenames[0]
                self.ui.cgmes_boundary_set_label.setText(self.current_boundary_set)

    def export(self) -> None:
        """
        The elements of the grid will be added with new idtags.
        This is useful in the case you want to compose a new grid from grids that are the same.
        :return:
        """
        cgmes_version = self.cgmes_version_dict[self.ui.cgmes_version_comboBox.currentText()]
        cgmes_profiles_txt = gf.get_checked_values(mdl=self.ui.cgmes_profiles_listView.model())
        cgmes_profiles = [self.cgmes_profiles_dict[e] for e in cgmes_profiles_txt]
        cgmes_one_file_per_profile = self.ui.cgmes_single_profile_per_file_checkBox.isChecked()
        cgmes_map_areas_like_raw = self.ui.cgmes_map_regions_like_raw_checkBox.isChecked()

        options = filedrv.FileSavingOptions(cgmes_boundary_set=self.current_boundary_set,
                                            sessions_data=self.sessions_data.get_save_data(),
                                            dictionary_of_json_files=dict(),
                                            cgmes_version=cgmes_version,
                                            cgmes_profiles=cgmes_profiles,
                                            cgmes_one_file_per_profile=cgmes_one_file_per_profile,
                                            cgmes_map_areas_like_raw=cgmes_map_areas_like_raw)

        # if the global file_name is empty, ask where to save
        fname = os.path.join(self.project_directory, self.app.ui.grid_name_line_edit.text())

        files_types = "CGMES (*.zip);;"
        filename, type_selected = QtWidgets.QFileDialog.getSaveFileName(self,
                                                                        'Export to CGMES',
                                                                        fname,
                                                                        files_types)

        if filename != '':

            # if the user did not enter the extension, add it automatically
            name, file_extension = os.path.splitext(filename)

            if file_extension == '':
                filename = name + '.zip'

            # we were able to compose the file correctly, now save it
            self.app.save_file_now(filename=filename,
                                   type_selected=type_selected,
                                   options=options)

            self.app.show_info_toast("CGMES export done!")
            self.close()




# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
import os
import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from PySide6.QtGui import Qt
from PySide6 import QtWidgets, QtCore
from PySide6.QtCore import Qt

from VeraGridEngine.IO.file_system import scripts_path
from VeraGrid.Gui.Main.SubClasses.io import IoMain
from VeraGrid.Gui.gui_functions import CustomFileSystemModel
import VeraGrid.Gui.gui_functions as gf
from VeraGrid.Gui.messages import error_msg, yes_no_question

from VeraGridEngine.IO.file_system import get_create_veragrid_folder
import VeraGridEngine as vg


class ScriptingMain(IoMain):
    """
    Diagrams Main
    """

    def __init__(self, parent=None):
        """

        @param parent:
        """

        # create main window
        IoMain.__init__(self, parent)

        self.add_console_vars()

        # scripts tree view --------------------------------------------------------------------------------------------
        self.python_fs_model = CustomFileSystemModel(root_path=scripts_path(), ext_filter=['*.py'])
        self.ui.sourceCodeTreeView.setModel(self.python_fs_model)
        self.ui.sourceCodeTreeView.setRootIndex(self.python_fs_model.index(scripts_path()))

        # actions ------------------------------------------------------------------------------------------------------
        self.ui.actionReset_console.triggered.connect(self.reset_console)

        # button clicks ------------------------------------------------------------------------------------------------
        self.ui.runSourceCodeButton.clicked.connect(self.run_source_code)
        self.ui.saveSourceCodeButton.clicked.connect(self.save_source_code)
        self.ui.clearSourceCodeButton.clicked.connect(self.clear_source_code)
        self.ui.clearConsoleButton.clicked.connect(self.reset_console)

        # double clicked -----------------------------------------------------------------------------------------------
        self.ui.sourceCodeTreeView.doubleClicked.connect(self.source_code_tree_clicked)

        # context menu
        self.ui.sourceCodeTreeView.customContextMenuRequested.connect(self.show_source_code_tree_context_menu)

        # Set context menu policy to CustomContextMenu
        self.ui.sourceCodeTreeView.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)

    def clear_console(self):
        """
        Clear console output
        """
        self.console.clear()

    def reset_console(self):
        """
        Reset console
        :return:
        """
        self.console.reset()
        self.add_console_vars()

    def print_console_help(self):

        self.console.append_output("")
        self.console.append_output('VeraGrid internal commands.\n')
        self.console.append_output('If a command is unavailable is because the study has not been executed yet.')

        self.console.append_output('\n\nclc():\tclear the console.')

        self.console.append_output('\n\nApp functions:')
        self.console.append_output('\tapp.new_project(): Clear all.')
        self.console.append_output('\tapp.open_file(): Prompt to load VeraGrid compatible file')
        self.console.append_output('\tapp.save_file(): Prompt to save VeraGrid file')
        self.console.append_output('\tapp.export_diagram(): Prompt to export the diagram in png.')
        self.console.append_output(
            '\tapp.create_schematic_from_api(): Create the schematic from the circuit information.')
        self.console.append_output(
            '\tapp.adjust_all_node_width(): Adjust the width of all the nodes according to their name.')
        self.console.append_output('\tapp.numerical_circuit: get compilation of the assets.')
        self.console.append_output('\tapp.islands: get compilation of the assets split into the topological islands.')

        self.console.append_output('\n\nCircuit functions:')
        self.console.append_output(
            '\tapp.circuit.plot_graph(): Plot a graph in a Matplotlib window. Call plt.show() after.')

        self.console.append_output('\n\nPower flow results:')
        self.console.append_output('\tapp.session.power_flow.voltage:\t the nodal voltages in per unit')
        self.console.append_output('\tapp.session.power_flow.current:\t the branch currents in per unit')
        self.console.append_output('\tapp.session.power_flow.loading:\t the branch loading in %')
        self.console.append_output('\tapp.session.power_flow.losses:\t the branch losses in per unit')
        self.console.append_output('\tapp.session.power_flow.power:\t the nodal power Injections in per unit')
        self.console.append_output(
            '\tapp.session.power_flow.Sf:\t the branch power Injections in per unit at the "from" side')
        self.console.append_output(
            '\tapp.session.power_flow.St:\t the branch power Injections in per unit at the "to" side')

        self.console.append_output('\n\nShort circuit results:')
        self.console.append_output('\tapp.session.short_circuit.voltage:\t the nodal voltages in per unit')
        self.console.append_output('\tapp.session.short_circuit.current:\t the branch currents in per unit')
        self.console.append_output('\tapp.session.short_circuit.loading:\t the branch loading in %')
        self.console.append_output('\tapp.session.short_circuit.losses:\t the branch losses in per unit')
        self.console.append_output('\tapp.session.short_circuit.power:\t the nodal power Injections in per unit')
        self.console.append_output(
            '\tapp.session.short_circuit.power_from:\t the branch power Injections in per unit at the "from" side')
        self.console.append_output(
            '\tapp.session.short_circuit.power_to:\t the branch power Injections in per unit at the "to" side')
        self.console.append_output(
            '\tapp.session.short_circuit.short_circuit_power:\t Short circuit power in MVA of the grid nodes')

        self.console.append_output('\n\nOptimal power flow results:')
        self.console.append_output('\tapp.session.optimal_power_flow.voltage:\t the nodal voltages angles in rad')
        self.console.append_output('\tapp.session.optimal_power_flow.load_shedding:\t the branch loading in %')
        self.console.append_output('\tapp.session.optimal_power_flow.losses:\t the branch losses in per unit')
        self.console.append_output('\tapp.session.optimal_power_flow.Sbus:\t the nodal power Injections in MW')
        self.console.append_output('\tapp.session.optimal_power_flow.Sf:\t the branch power Sf')

        self.console.append_output('\n\nTime series power flow results:')
        self.console.append_output(
            '\tapp.session.power_flow_ts.time:\t Profiles time index (pandas DateTimeIndex object)')
        self.console.append_output(
            '\tapp.session.power_flow_ts.load_profiles:\t Load profiles matrix (row: time, col: node)')
        self.console.append_output(
            '\tapp.session.power_flow_ts.gen_profiles:\t Generation profiles matrix (row: time, col: node)')
        self.console.append_output(
            '\tapp.session.power_flow_ts.voltages:\t nodal voltages results matrix (row: time, col: node)')
        self.console.append_output(
            '\tapp.session.power_flow_ts.currents:\t Branches currents results matrix (row: time, col: branch)')
        self.console.append_output(
            '\tapp.session.power_flow_ts.loadings:\t Branches loadings results matrix (row: time, col: branch)')
        self.console.append_output(
            '\tapp.session.power_flow_ts.losses:\t Branches losses results matrix (row: time, col: branch)')

        self.console.append_output('\n\nVoltage stability power flow results:')
        self.console.append_output(
            '\tapp.session.continuation_power_flow.voltage:\t Voltage values for every power multiplication factor.')
        self.console.append_output(
            '\tapp.session.continuation_power_flow.lambda:\t Value of power multiplication factor applied')
        self.console.append_output(
            '\tapp.session.continuation_power_flow.Sf:\t Power values for every power multiplication factor.')

        self.console.append_output('\n\nMonte Carlo power flow results:')
        self.console.append_output('\tapp.session.stochastic_power_flow.V_avg:\t nodal voltage average result.')
        self.console.append_output('\tapp.session.stochastic_power_flow.I_avg:\t branch current average result.')
        self.console.append_output('\tapp.session.stochastic_power_flow.Loading_avg:\t branch loading average result.')
        self.console.append_output('\tapp.session.stochastic_power_flow.Losses_avg:\t branch losses average result.')
        self.console.append_output(
            '\tapp.session.stochastic_power_flow.V_std:\t nodal voltage standard deviation result.')
        self.console.append_output(
            '\tapp.session.stochastic_power_flow.I_std:\t branch current standard deviation result.')
        self.console.append_output(
            '\tapp.session.stochastic_power_flow.Loading_std:\t branch loading standard deviation result.')
        self.console.append_output(
            '\tapp.session.stochastic_power_flow.Losses_std:\t branch losses standard deviation result.')
        self.console.append_output('\tapp.session.stochastic_power_flow.V_avg_series:\t nodal voltage average series.')
        self.console.append_output(
            '\tapp.session.stochastic_power_flow.V_std_series:\t branch current standard deviation series.')
        self.console.append_output(
            '\tapp.session.stochastic_power_flow.error_series:\t Monte Carlo error series (the convergence value).')
        self.console.append_output('The same for app.latin_hypercube_sampling')

    def add_console_vars(self):
        """
        Add vars to the console
        :return:
        """
        for elm in [self.console, self.code_editor]:
            elm.add_var("hlp", self.print_console_help)
            elm.add_var("np", np)
            elm.add_var("pd", pd)
            elm.add_var("plt", plt)
            elm.add_var("vg", vg)  # veragrid as a library
            elm.add_var('app', self)
            elm.add_var('circuit', self.circuit)
            elm.add_var('user_folder', get_create_veragrid_folder)

        self.console.add_var("clc", self.console.clear)

    def append_output(self, text: str):
        """
        Add some text to the output
        :param text: text to append
        """
        self.console.append_output(text)

    def run_source_code(self):
        """
        Run the source code in the IPython console
        """
        source_code = self.code_editor.toPlainText()

        if source_code[-1] != '\n':
            source_code += "\n"

        self.console.execute(source_code)
        self.console.append_output(">>> ")

    def source_code_tree_clicked(self, index):
        """
        On double click on a source code tree item, load the source code
        """
        pth = self.python_fs_model.filePath(index)

        if os.path.exists(pth):
            with open(pth, 'r') as f:
                txt = "\n".join(line.rstrip() for line in f)
                self.code_editor.setPlainText(txt)

            name = os.path.basename(pth)
            self.ui.sourceCodeNameLineEdit.setText(name.replace('.py', ''))
        else:
            error_msg(pth + ' does not exists :/', 'Open script')

    def clear_source_code(self):
        """
        Clear source code
        """
        ok = yes_no_question(text='Are you sure you want to clear source code?',
                             title='Clear Source Code')

        if ok:
            self.ui.sourceCodeNameLineEdit.setText("")
            self.code_editor.setPlainText("")

    def save_source_code(self):
        """
        Save the source code
        """
        name = self.ui.sourceCodeNameLineEdit.text().strip()

        if name != '':
            fname = name + '.py'
            pth = os.path.join(scripts_path(), fname)
            with open(pth, 'w') as f:
                f.write(self.code_editor.toPlainText())
        else:
            error_msg("Please enter a name for the script", title="Save script")

    def delete_source_code(self):
        """
        Delete the selected file
        """
        index = self.ui.sourceCodeTreeView.currentIndex()
        pth = self.python_fs_model.filePath(index)
        if os.path.exists(pth):
            ok = yes_no_question(text="Do you want to delete_with_dialogue {}?".format(pth),
                                 title="Delete source code file")

            if ok:
                os.remove(pth)
        else:
            error_msg(pth + ' does not exists :/', "Delete source code file")

    def show_source_code_tree_context_menu(self, pos: QtCore.QPoint):
        """
        Show source code tree view context menu
        :param pos: Relative click position
        """
        context_menu = QtWidgets.QMenu(parent=self.ui.diagramsListView)

        gf.add_menu_entry(menu=context_menu,
                          text="Delete",
                          icon_path=":/Icons/icons/delete_with_dialogue.png",
                          function_ptr=self.delete_source_code)

        # Convert global position to local position of the list widget
        mapped_pos = self.ui.sourceCodeTreeView.viewport().mapToGlobal(pos)
        context_menu.exec(mapped_pos)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
import numpy as np
from typing import List
from PySide6.QtCore import Qt, QModelIndex
from PySide6.QtWidgets import QTableView, QVBoxLayout, QPushButton, QHBoxLayout, QDialog, QSpacerItem, QSizePolicy
from VeraGrid.Gui.general_dialogues import ArrayTableModel
from VeraGridEngine.Devices.Injections.controllable_shunt import ControllableShunt


class ControllableShuntArray(ArrayTableModel):

    def __init__(self, data: List[np.ndarray], headers: List[str], dtypes: List[np.dtype]):
        ArrayTableModel.__init__(self, data=data, headers=headers)
        self.dtypes = dtypes

    def setData(self, index: QModelIndex, value: float, role=Qt.ItemDataRole.EditRole):
        """

        :param index:
        :param value:
        :param role:
        :return:
        """
        if not index.isValid():
            return False

        if role == Qt.ItemDataRole.EditRole:
            row = index.row()
            column = index.column()
            try:
                value = float(value)
            except ValueError:
                return False

            tpe = self.dtypes[column]
            self._data[column][row] = tpe(value)

            self.dataChanged.emit(index, index, [Qt.ItemDataRole.DisplayRole, Qt.ItemDataRole.EditRole])
            return True

        return False


class ControllableShuntEditor(QDialog):
    """
    ArrayEditor
    """

    def __init__(self, api_object: ControllableShunt):
        QDialog.__init__(self)

        self.setWindowTitle("Controllable shunt editor")

        self.api_object = api_object
        self.model = ControllableShuntArray(data=[self.api_object.active_steps,
                                                  self.api_object.g_steps,
                                                  self.api_object.b_steps],
                                            headers=["Active", "G steps (MW)", "B steps (MVAr)"],
                                            dtypes=[bool, float, float])

        self.table_view = QTableView()
        self.table_view.setModel(self.model)

        self.add_button = QPushButton("Add")
        self.delete_button = QPushButton("Delete")
        self.done_button = QPushButton("Done")

        self.add_button.clicked.connect(self.add_row)
        self.delete_button.clicked.connect(self.delete_row)
        self.done_button.clicked.connect(self.accept_click)

        layout = QVBoxLayout()
        layout.addWidget(self.table_view)

        button_layout = QHBoxLayout()
        button_layout.addWidget(self.add_button)
        button_layout.addWidget(self.delete_button)
        button_layout.addItem(QSpacerItem(40, 20, QSizePolicy.Policy.Maximum))
        button_layout.addWidget(self.done_button)

        layout.addLayout(button_layout)

        self.setLayout(layout)

    def get_active_steps(self) -> np.ndarray:
        """

        :return:
        """
        return self.model.get_data()[0]

    def get_g_steps(self) -> np.ndarray:
        """

        :return:
        """
        return self.model.get_data()[1]

    def get_b_steps(self) -> np.ndarray:
        """

        :return:
        """
        return self.model.get_data()[2]

    def add_row(self):
        """
        Add row
        """
        row_count = self.model.rowCount()
        self.model.insertRows(row_count, 1)

    def delete_row(self):
        """
        Delete the selected rows
        """
        selected_indexes = self.table_view.selectionModel().selectedIndexes()

        rows = list({index.row() for index in selected_indexes})
        rows.sort(reverse=True)
        for r in rows:
            self.model.removeRows(position=r, rows=1)

    def accept_click(self):
        """
        Modify and exit
        """
        self.api_object.active_steps = self.get_active_steps()
        self.api_object.g_steps = self.get_g_steps()
        self.api_object.Gmax = self.api_object.g_steps.max()
        self.api_object.Gmin = self.api_object.g_steps.min()
        self.api_object.b_steps = self.get_b_steps()
        self.api_object.Bmax = self.api_object.b_steps.max()
        self.api_object.Bmin = self.api_object.b_steps.min()
        self.accept()



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'sync_gui.ui'
##
## Created by: Qt User Interface Compiler version 6.9.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QAbstractItemView, QApplication, QDialog, QFrame,
    QHBoxLayout, QHeaderView, QPushButton, QSizePolicy,
    QSpacerItem, QTreeView, QVBoxLayout, QWidget)
from VeraGrid.Gui.Icons.icons_rc import *

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        if not Dialog.objectName():
            Dialog.setObjectName(u"Dialog")
        Dialog.resize(941, 590)
        self.verticalLayout = QVBoxLayout(Dialog)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(1, 1, 1, 1)
        self.frame = QFrame(Dialog)
        self.frame.setObjectName(u"frame")
        self.frame.setFrameShape(QFrame.NoFrame)
        self.frame.setFrameShadow(QFrame.Raised)
        self.verticalLayout_2 = QVBoxLayout(self.frame)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.verticalLayout_2.setContentsMargins(-1, -1, -1, 0)
        self.treeView = QTreeView(self.frame)
        self.treeView.setObjectName(u"treeView")
        self.treeView.setAlternatingRowColors(True)
        self.treeView.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.treeView.setAnimated(True)

        self.verticalLayout_2.addWidget(self.treeView)


        self.verticalLayout.addWidget(self.frame)

        self.frame_2 = QFrame(Dialog)
        self.frame_2.setObjectName(u"frame_2")
        self.frame_2.setFrameShape(QFrame.NoFrame)
        self.frame_2.setFrameShadow(QFrame.Raised)
        self.horizontalLayout = QHBoxLayout(self.frame_2)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalLayout.setContentsMargins(-1, 0, -1, -1)
        self.accept_selected_pushButton = QPushButton(self.frame_2)
        self.accept_selected_pushButton.setObjectName(u"accept_selected_pushButton")
        icon = QIcon()
        icon.addFile(u":/Icons/icons/accept.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.accept_selected_pushButton.setIcon(icon)

        self.horizontalLayout.addWidget(self.accept_selected_pushButton)

        self.reject_selected_pushButton = QPushButton(self.frame_2)
        self.reject_selected_pushButton.setObjectName(u"reject_selected_pushButton")
        icon1 = QIcon()
        icon1.addFile(u":/Icons/icons/delete.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.reject_selected_pushButton.setIcon(icon1)

        self.horizontalLayout.addWidget(self.reject_selected_pushButton)

        self.horizontalSpacer = QSpacerItem(632, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout.addItem(self.horizontalSpacer)

        self.doit_pushButton = QPushButton(self.frame_2)
        self.doit_pushButton.setObjectName(u"doit_pushButton")
        icon2 = QIcon()
        icon2.addFile(u":/Icons/icons/accept2.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.doit_pushButton.setIcon(icon2)

        self.horizontalLayout.addWidget(self.doit_pushButton)


        self.verticalLayout.addWidget(self.frame_2)


        self.retranslateUi(Dialog)

        QMetaObject.connectSlotsByName(Dialog)
    # setupUi

    def retranslateUi(self, Dialog):
        Dialog.setWindowTitle(QCoreApplication.translate("Dialog", u"Dialog", None))
#if QT_CONFIG(tooltip)
        self.accept_selected_pushButton.setToolTip(QCoreApplication.translate("Dialog", u"Accept selected", None))
#endif // QT_CONFIG(tooltip)
        self.accept_selected_pushButton.setText("")
#if QT_CONFIG(tooltip)
        self.reject_selected_pushButton.setToolTip(QCoreApplication.translate("Dialog", u"Reject selected changes", None))
#endif // QT_CONFIG(tooltip)
        self.reject_selected_pushButton.setText("")
#if QT_CONFIG(tooltip)
        self.doit_pushButton.setToolTip(QCoreApplication.translate("Dialog", u"Process all changes as especified", None))
#endif // QT_CONFIG(tooltip)
        self.doit_pushButton.setText(QCoreApplication.translate("Dialog", u"Do it", None))
    # retranslateUi




# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
from __future__ import annotations
from uuid import uuid4
import pandas as pd
from PySide6.QtCore import QThread, Signal
from typing import Dict, Union, List, Tuple, Any, Generator
from collections.abc import Callable
from warnings import warn

from VeraGridEngine.Simulations.ATC.available_transfer_capacity_driver import (AvailableTransferCapacityDriver,
                                                                               AvailableTransferCapacityResults)
from VeraGridEngine.Simulations.ATC.available_transfer_capacity_ts_driver import (
    AvailableTransferCapacityTimeSeriesDriver, AvailableTransferCapacityTimeSeriesResults
)
from VeraGridEngine.Simulations.ContingencyAnalysis.contingency_analysis_driver import (ContingencyAnalysisDriver,
                                                                                        ContingencyAnalysisResults)
from VeraGridEngine.Simulations.ContingencyAnalysis.contingency_analysis_ts_driver import (
    ContingencyAnalysisTimeSeriesDriver, ContingencyAnalysisTimeSeriesResults
)
from VeraGridEngine.Simulations.ContinuationPowerFlow.continuation_power_flow_driver import (
    ContinuationPowerFlowDriver, ContinuationPowerFlowResults
)
from VeraGridEngine.Simulations.LinearFactors.linear_analysis_driver import LinearAnalysisDriver, LinearAnalysisResults
from VeraGridEngine.Simulations.LinearFactors.linear_analysis_ts_driver import (LinearAnalysisTimeSeriesDriver,
                                                                                LinearAnalysisTimeSeriesResults)
from VeraGridEngine.Simulations.OPF.opf_driver import OptimalPowerFlowDriver, OptimalPowerFlowResults
from VeraGridEngine.Simulations.OPF.opf_ts_driver import (OptimalPowerFlowTimeSeriesDriver,
                                                          OptimalPowerFlowTimeSeriesResults)
from VeraGridEngine.Simulations.PowerFlow.power_flow_driver import PowerFlowDriver, PowerFlowResults
from VeraGridEngine.Simulations.PowerFlow.power_flow_driver_3ph import PowerFlowDriver3Ph, PowerFlowResults3Ph
from VeraGridEngine.Simulations.PowerFlow.power_flow_ts_driver import (PowerFlowTimeSeriesDriver,
                                                                       PowerFlowTimeSeriesResults)
from VeraGridEngine.Simulations.ShortCircuitStudies.short_circuit_driver import ShortCircuitDriver, ShortCircuitResults
from VeraGridEngine.Simulations.Stochastic.stochastic_power_flow_driver import (StochasticPowerFlowDriver,
                                                                                StochasticPowerFlowResults)
from VeraGridEngine.Simulations.Clustering.clustering_driver import ClusteringDriver, ClusteringResults
from VeraGridEngine.Simulations.Reliability.blackout_driver import CascadingResults, CascadingDriver
from VeraGridEngine.Simulations.InputsAnalysis.inputs_analysis_driver import InputsAnalysisResults, InputsAnalysisDriver
from VeraGridEngine.Simulations.InvestmentsEvaluation.investments_evaluation_driver import (InvestmentsEvaluationDriver,
                                                                                            InvestmentsEvaluationResults)
from VeraGridEngine.Simulations.SigmaAnalysis.sigma_analysis_driver import SigmaAnalysisResults
from VeraGridEngine.Simulations.NTC.ntc_driver import (OptimalNetTransferCapacityResults,
                                                       OptimalNetTransferCapacityDriver)
from VeraGridEngine.Simulations.NodalCapacity.nodal_capacity_ts_driver import (NodalCapacityTimeSeriesDriver,
                                                                               NodalCapacityTimeSeriesResults)
from VeraGridEngine.Simulations.Reliability.reliability_driver import ReliabilityStudyDriver, ReliabilityResults
from VeraGridEngine.Simulations.StateEstimation.state_stimation_driver import (StateEstimationDriver,
                                                                               StateEstimationResults)
from VeraGridEngine.Simulations.SmallSignalStability.small_signal_driver import (SmallSignalStabilityDriver,
                                                                                 SmallSignalStabilityResults)
from VeraGridEngine.Simulations.Rms.rms_driver import RmsSimulationDriver, RmsResults
from VeraGridEngine.Simulations.Topology.node_groups_driver import NodeGroupsDriver
from VeraGridEngine.Simulations.driver_template import DriverTemplate, DriverToSave
from VeraGridEngine.Devices.multi_circuit import MultiCircuit
from VeraGridEngine.enumerations import ResultTypes, SimulationTypes
from VeraGridEngine.Simulations.driver_handler import create_driver
from VeraGridEngine.Simulations.types import DRIVER_OBJECTS, RESULTS_OBJECTS
from VeraGridEngine.basic_structures import Logger
from VeraGrid.Gui.results_model import ResultsModel


class GcThread(QThread):
    """
    Generic VeraGrid Thread
    this is used as a template for a Qt Thread
    """
    progress_signal = Signal(float)
    progress_text = Signal(str)
    done_signal = Signal()

    def __init__(self, driver: DriverTemplate):
        QThread.__init__(self)

        # assign the driver and set the driver's reporting functions
        self.driver = driver
        self.driver.progress_signal = self.progress_signal
        self.driver.progress_text = self.progress_text
        self.driver.done_signal = self.done_signal
        self.tpe = driver.tpe

        self.results = None

        self.elapsed = 0

        self.logger = Logger()

        self.__cancel__ = False

    def get_steps(self) -> List[Any]:
        """

        :return:
        """
        return list()

    def run(self) -> None:
        """
        Run driver
        """
        self.progress_signal.emit(0.0)

        self.driver.run()

        self.progress_signal.emit(0.0)
        if self.__cancel__:
            self.progress_text.emit('Cancelled!')
        else:
            self.progress_text.emit('Done!')
        self.done_signal.emit()

    def cancel(self) -> None:
        """
        Cancel the simulation
        """
        self.__cancel__ = True
        self.driver.__cancel__ = True


class SimulationSession:
    """
    The simulation session is the simulation manager
    it serves to orchestrate the threads and to store the drivers and their results
    """

    def __init__(self, name: str = 'Session', idtag: str = None):
        """
        Constructor
        :param name: Session name
        :param idtag: Unique identifier
        """
        self.idtag: str = uuid4().hex if idtag is None else idtag

        # name of the session
        self.name: str = name

        # dictionary of drivers
        self.drivers: Dict[SimulationTypes, DRIVER_OBJECTS] = dict()
        self.threads: Dict[SimulationTypes, GcThread] = dict()

    def __str__(self):
        return self.name

    def clear(self) -> None:
        """
        Delete all the drivers
        """
        self.drivers = dict()

    def register(self, driver: DRIVER_OBJECTS):
        """
        Register driver
        :param driver: driver to register (must have a tpe variable in it)
        """
        # register
        self.drivers[driver.tpe] = driver

    def get_save_data(self) -> List[DriverToSave]:
        """
        Get data to be saved
        :return: List[DriverToSave]
        """
        data = list()
        for tpe, drv in self.drivers.items():
            data.append(drv.get_save_data())
        return data

    def run(self,
            driver: DRIVER_OBJECTS,
            post_func: Union[None, Callable] = None,
            prog_func: Union[None, Callable] = None,
            text_func: Union[None, Callable] = None):
        """
        Register driver
        :param driver: driver to register (must have a tpe variable in it)
        :param post_func: Function to run after it is done
        :param prog_func: Function to display the progress
        :param text_func: Function to display text
        """

        # create a process
        thr = GcThread(driver)
        thr.progress_signal.connect(prog_func)
        thr.progress_text.connect(text_func)
        thr.done_signal.connect(post_func)

        # check and kill
        if driver.tpe in self.drivers.keys():
            del self.drivers[driver.tpe]
            if self.threads[driver.tpe].isRunning():
                self.threads[driver.tpe].terminate()
            del self.threads[driver.tpe]

        # register
        self.drivers[driver.tpe] = driver
        self.threads[driver.tpe] = thr

        # run!
        thr.start()

    def register_driver(self, driver: DRIVER_OBJECTS):
        """
        Register driver
        :param driver:
        :return:
        """
        self.drivers[driver.tpe] = driver

    def get_available_drivers(self):
        """
        Get a list of the available driver objects
        :return: List[Driver]
        """
        return [drv for driver_type, drv in self.drivers.items() if drv is not None]

    def drivers_results_iter(self) -> Generator[Tuple[DRIVER_OBJECTS | None, RESULTS_OBJECTS | None]]:
        """
        Iterator returning driver and result types
        :return: driver, result (both can be None)
        """
        for driver_type, drv in self.drivers.items():
            if hasattr(drv, 'results'):
                yield drv, drv.results
            else:
                yield drv, None

    def exists(self, driver_type: SimulationTypes):
        """
        Get the results of the driver
        :param driver_type: driver type to look for
        :return: True / False
        """
        return driver_type in self.drivers.keys()

    def get_driver_results(
            self,
            driver_type: SimulationTypes) -> Tuple[Union[None, DRIVER_OBJECTS], Union[None, RESULTS_OBJECTS]]:
        """
        Get the results of the driver
        :param driver_type: driver type
        :return: driver, results (None, None if not found)
        """

        drv: DRIVER_OBJECTS = self.drivers.get(driver_type, None)

        if drv is not None:
            if hasattr(drv, 'results'):
                return drv, drv.results
            else:
                return drv, None
        else:
            return None, None

    def get_results(self, driver_type: SimulationTypes) -> RESULTS_OBJECTS | None:
        """
        Get the results of the driver
        :param driver_type: driver type
        :return: driver, results (None, None if not found)
        """
        if driver_type in self.drivers.keys():
            drv = self.drivers[driver_type]
            if hasattr(drv, 'results'):
                return drv.results
            else:
                return None
        else:
            return None

    def delete_driver(self, driver_type: SimulationTypes) -> None:
        """
        Get the results of the driver
        :param driver_type: driver type to delete_with_dialogue
        """
        if driver_type in self.drivers.keys():
            del self.drivers[driver_type]

    def delete_driver_by_name(self, study_name: str) -> None:
        """
        Delete the driver by it's name
        :param study_name: driver name
        """
        for driver_type, drv in self.drivers.items():
            if study_name == drv.tpe.value:
                del self.drivers[driver_type]
                return
            if study_name == drv.name:
                del self.drivers[driver_type]
                return

    def get_driver_by_name(self,
                           study_name: str) -> Union[DRIVER_OBJECTS, None]:
        """
        Get the driver by it's name
        :param study_name: driver name
        """
        for driver_type, drv in self.drivers.items():
            if study_name == drv.name:
                return self.drivers[driver_type]
            if study_name == drv.tpe.value:
                return self.drivers[driver_type]
        return None

    def get_results_model_by_name(self,
                                  study_name: str,
                                  study_type: ResultTypes) -> Union[ResultsModel, None]:
        """
        Get the results model given the study name and study type
        :param study_name: name of the study
        :param study_type: name of the study type
        :return: ResultsModel instance or None if not found
        """
        for driver_type, drv in self.drivers.items():
            if study_name == drv.tpe.value or study_name == drv.name:
                if drv.results is not None:
                    return ResultsModel(drv.results.mdl(result_type=study_type))
                else:
                    print('There seem to be no results :(')
                    return None

        return None

    def register_driver_from_disk_data(self,
                                       grid: MultiCircuit,
                                       study_name: str,
                                       data_dict: Dict[str, pd.DataFrame]) -> Logger:
        """
        Create driver with the results
        :param grid: MultiCircuit instance
        :param study_name: name of the study (i.e. Power Flow)
        :param data_dict: dictionary of data coming from the file
        """
        logger = Logger()

        time_indices = data_dict.get('time_indices', grid.get_all_time_indices())

        driver_tpe = SimulationTypes(study_name)

        drv: DRIVER_OBJECTS | None = create_driver(
            grid=grid,
            driver_tpe=driver_tpe,
            time_indices=time_indices
        )

        if drv is not None:

            # fill in the saved results
            drv.results.parse_saved_data(grid=grid, data_dict=data_dict, logger=logger)

            # perform whatever operations are needed after loading
            drv.results.consolidate_after_loading()

            # parse the logger if available
            logger_data = data_dict.get('logger', None)
            if logger_data is not None:
                drv.logger.parse_df(df=logger_data)

            # register the driver
            self.register(drv)
        else:
            warn(f"Session {study_name} not implemented for disk retrieval :/")

        return logger

    def is_this_running(self, sim_tpe: SimulationTypes) -> bool:
        """
        Check if a simulation type is running
        :param sim_tpe: SimulationTypes
        :return: True / False
        """
        for driver_type, drv in self.threads.items():
            if drv is not None:
                if drv.isRunning():
                    if driver_type == sim_tpe:
                        return True
        return False

    def is_anything_running(self) -> bool:
        """
        Check if anything is running
        :return: True / False
        """

        for driver_type, drv in self.threads.items():
            if drv is not None:
                if drv.isRunning():
                    return True
        return False

    @property
    def clustering(self) -> Tuple[ClusteringDriver, ClusteringResults]:
        """

        :return:
        """
        drv, results = self.get_driver_results(SimulationTypes.ClusteringAnalysis_run)
        return drv, results

    @property
    def power_flow(self) -> Tuple[PowerFlowDriver, PowerFlowResults]:
        """

        :return:
        """
        drv, results = self.get_driver_results(SimulationTypes.PowerFlow_run)
        return drv, results

    @property
    def power_flow_3ph(self) -> Tuple[PowerFlowDriver3Ph, PowerFlowResults3Ph]:
        """

        :return:
        """
        drv, results = self.get_driver_results(SimulationTypes.PowerFlow3ph_run)
        return drv, results

    @property
    def state_estimation(self) -> Tuple[StateEstimationDriver, StateEstimationResults]:
        """

        :return:
        """
        drv, results = self.get_driver_results(SimulationTypes.StateEstimation_run)
        return drv, results

    @property
    def power_flow_ts(self) -> Tuple[PowerFlowTimeSeriesDriver, PowerFlowTimeSeriesResults]:
        """

        :return:
        """
        drv, results = self.get_driver_results(SimulationTypes.PowerFlowTimeSeries_run)
        return drv, results

    @property
    def optimal_power_flow(self) -> Tuple[OptimalPowerFlowDriver, OptimalPowerFlowResults]:
        """

        :return:
        """
        drv, results = self.get_driver_results(SimulationTypes.OPF_run)
        return drv, results

    @property
    def optimal_power_flow_ts(self) -> Tuple[OptimalPowerFlowTimeSeriesDriver, OptimalPowerFlowTimeSeriesResults]:
        """

        :return:
        """
        drv, results = self.get_driver_results(SimulationTypes.OPFTimeSeries_run)
        return drv, results

    @property
    def short_circuit(self) -> Tuple[ShortCircuitDriver, ShortCircuitResults]:
        """

        :return:
        """
        drv, results = self.get_driver_results(SimulationTypes.ShortCircuit_run)
        return drv, results

    @property
    def linear_power_flow(self) -> Tuple[LinearAnalysisDriver, LinearAnalysisResults]:
        """

        :return:
        """
        drv, results = self.get_driver_results(SimulationTypes.LinearAnalysis_run)
        return drv, results

    @property
    def linear_power_flow_ts(self) -> Tuple[LinearAnalysisTimeSeriesDriver, LinearAnalysisTimeSeriesResults]:
        """

        :return:
        """
        drv, results = self.get_driver_results(SimulationTypes.LinearAnalysis_TS_run)
        return drv, results

    @property
    def contingency(self) -> Tuple[ContingencyAnalysisDriver, ContingencyAnalysisResults]:
        """

        :return:
        """
        drv, results = self.get_driver_results(SimulationTypes.ContingencyAnalysis_run)
        return drv, results

    @property
    def contingency_ts(self) -> Tuple[ContingencyAnalysisTimeSeriesDriver, ContingencyAnalysisTimeSeriesResults]:
        """

        :return:
        """
        drv, results = self.get_driver_results(SimulationTypes.ContingencyAnalysisTS_run)
        return drv, results

    @property
    def continuation_power_flow(self) -> Tuple[ContinuationPowerFlowDriver, ContinuationPowerFlowResults]:
        """

        :return:
        """
        drv, results = self.get_driver_results(SimulationTypes.ContinuationPowerFlow_run)
        return drv, results

    @property
    def net_transfer_capacity(self) -> Tuple[AvailableTransferCapacityDriver, AvailableTransferCapacityResults]:
        """

        :return:
        """
        drv, results = self.get_driver_results(SimulationTypes.NetTransferCapacity_run)
        return drv, results

    @property
    def net_transfer_capacity_ts(self) -> Tuple[AvailableTransferCapacityTimeSeriesDriver,
    AvailableTransferCapacityTimeSeriesResults]:
        """

        :return:
        """
        drv, results = self.get_driver_results(SimulationTypes.NetTransferCapacityTS_run)
        return drv, results

    @property
    def optimal_net_transfer_capacity(self) -> Tuple[
        OptimalNetTransferCapacityDriver, OptimalNetTransferCapacityResults]:
        """

        :return:
        """
        drv, results = self.get_driver_results(SimulationTypes.OPF_NTC_run)
        return drv, results

    @property
    def optimal_net_transfer_capacity_ts(self) -> Tuple[
        OptimalNetTransferCapacityDriver, OptimalNetTransferCapacityResults]:
        """

        :return:
        """
        drv, results = self.get_driver_results(SimulationTypes.OPF_NTC_TS_run)
        return drv, results

    @property
    def nodal_capacity_optimization_ts(self) -> Tuple[NodalCapacityTimeSeriesDriver, NodalCapacityTimeSeriesResults]:
        """

        :return:
        """
        drv, results = self.get_driver_results(SimulationTypes.NodalCapacityTimeSeries_run)
        return drv, results

    @property
    def reliability_analysis(self) -> Tuple[ReliabilityStudyDriver, ReliabilityResults]:
        """

        :return:
        """
        drv, results = self.get_driver_results(SimulationTypes.Reliability_run)
        return drv, results

    @property
    def rms_dynamic_simulation(self) -> Tuple[RmsSimulationDriver, RmsResults]:
        """

        :return:
        """
        drv, results = self.get_driver_results(SimulationTypes.RmsDynamic_run)
        return drv, results

    @property
    def stochastic_power_flow(self) -> Tuple[StochasticPowerFlowDriver, StochasticPowerFlowResults]:
        """

        :return:
        """
        drv, results = self.get_driver_results(SimulationTypes.StochasticPowerFlow)
        return drv, results

    @property
    def sigma_analysis(self) -> Tuple[ShortCircuitDriver, SigmaAnalysisResults]:
        """

        :return:
        """
        drv, results = self.get_driver_results(SimulationTypes.SigmaAnalysis_run)
        return drv, results

    @property
    def cascade(self) -> Tuple[CascadingDriver, CascadingResults]:
        """

        :return:
        """
        drv, results = self.get_driver_results(SimulationTypes.Cascade_run)
        return drv, results

    @property
    def inputs_analysis(self) -> Tuple[InputsAnalysisDriver, InputsAnalysisResults]:
        """

        :return:
        """
        drv, results = self.get_driver_results(SimulationTypes.InputsAnalysis_run)
        return drv, results

    @property
    def investments_evaluation(self) -> Tuple[InvestmentsEvaluationDriver, InvestmentsEvaluationResults]:
        """

        :return:
        """
        drv, results = self.get_driver_results(SimulationTypes.InvestmentsEvaluation_run)
        return drv, results

    @property
    def node_groups_driver(self) -> Tuple[NodeGroupsDriver, None]:
        """

        :return:
        """
        drv, results = self.get_driver_results(SimulationTypes.NodeGrouping_run)
        return drv, None

    @property
    def small_signal_stability_simulation(self) -> Tuple[SmallSignalStabilityDriver, SmallSignalStabilityResults]:
        """

        :return:
        """
        drv, results = self.get_driver_results(SimulationTypes.SmallSignal_run)
        return drv, results
